# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pkgutil
import shutil
import subprocess
import sys
import textwrap
from types import ModuleType
from typing import Callable
from unittest import mock

import fixtures
import jsonschema

import snapcraft
from snapcraft.internal import elf
from snapcraft.plugins._plugin_finder import get_plugin_for_base
from tests.file_utils import get_snapcraft_path


class FakeProjectOptions(fixtures.Fixture):
    def __init__(self, **kwargs):
        self._kwargs = dict(
            arch_triplet=kwargs.pop("arch_triplet", "x86_64-gnu-linux"),
            parts_dir=kwargs.pop("parts_dir", "parts"),
            stage_dir=kwargs.pop("stage_dir", "stage"),
            prime_dir=kwargs.pop("prime_dir", "prime"),
            parallel_build_count=kwargs.pop("parallel_build_count", "1"),
        )
        if kwargs:
            raise NotImplementedError(
                "Handling of {!r} is not implemented".format(kwargs.keys())
            )

    def setUp(self):
        super().setUp()

        patcher = mock.patch("snapcraft.project.Project")
        patcher.start()
        self.addCleanup(patcher.stop)

        # Special handling is required as ProjectOptions attributes are
        # handled with the @property decorator.
        project_options_t = type(snapcraft.project.Project.return_value)
        for key in self._kwargs:
            setattr(project_options_t, key, self._kwargs[key])


class FakeMetadataExtractor(fixtures.Fixture):
    """Dynamically generate a new module containing the provided extractor"""

    def __init__(
        self,
        extractor_name: str,
        extractor: Callable[[str], snapcraft.extractors.ExtractedMetadata],
        exported_name="extract",
    ) -> None:
        super().__init__()
        self._extractor_name = extractor_name
        self._exported_name = exported_name
        self._import_name = "snapcraft.extractors.{}".format(extractor_name)
        self._extractor = extractor

    def _setUp(self) -> None:
        extractor_module = ModuleType(self._import_name)
        setattr(extractor_module, self._exported_name, self._extractor)
        sys.modules[self._import_name] = extractor_module
        self.addCleanup(self._remove_module)

        real_iter_modules = pkgutil.iter_modules

        def _fake_iter_modules(path):
            if path == snapcraft.extractors.__path__:
                yield None, self._extractor_name, False
            else:
                yield real_iter_modules(path)

        patcher = mock.patch("pkgutil.iter_modules", new=_fake_iter_modules)
        patcher.start()
        self.addCleanup(patcher.stop)

    def _remove_module(self) -> None:
        del sys.modules[self._import_name]


class FakePlugin(fixtures.Fixture):
    """Dynamically generate a new module containing the provided plugin"""

    def __init__(self, plugin_name, plugin_class):
        super().__init__()
        self._plugin_name = plugin_name
        self._plugin_class = plugin_class

    def _setUp(self):
        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.plugins.get_plugin_for_base", side_effect=self.get_plugin
            )
        )

    def get_plugin(self, plugin_name, *, build_base):
        if plugin_name == self._plugin_name:
            plugin_class = self._plugin_class
        else:
            plugin_class = get_plugin_for_base(self._plugin_name, build_base=build_base)

        return plugin_class


def _fake_elffile_extract_attributes(self):
    name = os.path.basename(self.path)

    self.arch = ("ELFCLASS64", "ELFDATA2LSB", "EM_X86_64")
    self.build_id = "build-id-{}".format(name)

    if name in [
        "fake_elf-2.26",
        "fake_elf-bad-ldd",
        "fake_elf-with-core-libs",
        "fake_elf-with-missing-libs",
        "fake_elf-bad-patchelf",
        "fake_elf-with-host-libraries",
    ]:
        glibc = elf.NeededLibrary(name="libc.so.6")
        glibc.add_version("GLIBC_2.2.5")
        glibc.add_version("GLIBC_2.26")

        self.interp = "/lib64/ld-linux-x86-64.so.2"
        self.soname = ""
        self.versions = set()
        self.needed = {glibc.name: glibc}
        self.execstack_set = False
        self.is_dynamic = True
        self.has_debug_info = False

    elif name == "fake_elf-2.23":
        glibc = elf.NeededLibrary(name="libc.so.6")
        glibc.add_version("GLIBC_2.2.5")
        glibc.add_version("GLIBC_2.23")

        self.interp = "/lib64/ld-linux-x86-64.so.2"
        self.soname = ""
        self.versions = set()
        self.needed = {glibc.name: glibc}
        self.execstack_set = False
        self.is_dynamic = True
        self.has_debug_info = False

    elif name == "fake_elf-1.1":
        glibc = elf.NeededLibrary(name="libc.so.6")
        glibc.add_version("GLIBC_1.1")
        glibc.add_version("GLIBC_0.1")

        self.interp = "/lib64/ld-linux-x86-64.so.2"
        self.soname = ""
        self.versions = set()
        self.needed = {glibc.name: glibc}
        self.execstack_set = False
        self.is_dynamic = True
        self.has_debug_info = False

    elif name == "fake_elf-static":
        self.interp = "/lib64/ld-linux-x86-64.so.2"
        self.soname = ""
        self.versions = set()
        self.needed = dict()
        self.execstack_set = False
        self.is_dynamic = False
        self.has_debug_info = False

    elif name == "fake_elf-shared-object":
        openssl = elf.NeededLibrary(name="libssl.so.1.0.0")
        openssl.add_version("OPENSSL_1.0.0")

        self.interp = ""
        self.soname = "libfake_elf.so.0"
        self.versions = set()
        self.needed = {openssl.name: openssl}
        self.execstack_set = False
        self.is_dynamic = True
        self.has_debug_info = False

    elif name == "fake_elf-with-execstack":
        glibc = elf.NeededLibrary(name="libc.so.6")
        glibc.add_version("GLIBC_2.23")

        self.interp = "/lib64/ld-linux-x86-64.so.2"
        self.soname = ""
        self.versions = set()
        self.needed = {glibc.name: glibc}
        self.execstack_set = True
        self.is_dynamic = True
        self.has_debug_info = False

    elif name == "fake_elf-with-bad-execstack":
        glibc = elf.NeededLibrary(name="libc.so.6")
        glibc.add_version("GLIBC_2.23")

        self.interp = "/lib64/ld-linux-x86-64.so.2"
        self.soname = ""
        self.versions = set()
        self.needed = {glibc.name: glibc}
        self.execstack_set = True
        self.is_dynamic = True
        self.has_debug_info = False

    elif name == "libc.so.6":
        self.interp = ""
        self.soname = "libc.so.6"
        self.versions = {"libc.so.6", "GLIBC_2.2.5", "GLIBC_2.23", "GLIBC_2.26"}
        self.needed = {}
        self.execstack_set = False
        self.is_dynamic = True
        self.has_debug_info = False

    elif name == "libssl.so.1.0.0":
        self.interp = ""
        self.soname = "libssl.so.1.0.0"
        self.versions = {"libssl.so.1.0.0", "OPENSSL_1.0.0"}
        self.needed = {}
        self.execstack_set = False
        self.is_dynamic = True
        self.has_debug_info = False

    else:
        self.interp = ""
        self.soname = ""
        self.versions = set()
        self.needed = {}
        self.execstack_set = False
        self.is_dynamic = True
        self.has_debug_info = False


class FakeElf(fixtures.Fixture):
    def __getitem__(self, item):
        return self._elf_files[item]

    def __init__(self, *, root_path, patchelf_version="0.10"):
        super().__init__()

        self.root_path = root_path
        self.core_base_path = None
        self._patchelf_version = patchelf_version

    def _setUp(self):
        super()._setUp()

        self.core_base_path = self.useFixture(fixtures.TempDir()).path

        binaries_path = os.path.join(get_snapcraft_path(), "tests", "bin", "elf")

        new_binaries_path = self.useFixture(fixtures.TempDir()).path
        current_path = os.environ.get("PATH")
        new_path = "{}:{}".format(new_binaries_path, current_path)
        self.useFixture(fixtures.EnvironmentVariable("PATH", new_path))

        # Copy strip
        for f in ["strip", "execstack"]:
            shutil.copy(
                os.path.join(binaries_path, f), os.path.join(new_binaries_path, f)
            )
            os.chmod(os.path.join(new_binaries_path, f), 0o755)

        # Some values in ldd need to be set with core_path
        with open(os.path.join(binaries_path, "ldd")) as rf:
            with open(os.path.join(new_binaries_path, "ldd"), "w") as wf:
                for line in rf.readlines():
                    wf.write(line.replace("{CORE_PATH}", self.core_base_path))
        os.chmod(os.path.join(new_binaries_path, "ldd"), 0o755)

        # Some values in ldd need to be set with core_path
        self.patchelf_path = os.path.join(new_binaries_path, "patchelf")
        with open(os.path.join(binaries_path, "patchelf")) as rf:
            with open(self.patchelf_path, "w") as wf:
                for line in rf.readlines():
                    wf.write(line.replace("{VERSION}", self._patchelf_version))
        os.chmod(os.path.join(new_binaries_path, "patchelf"), 0o755)

        patcher = mock.patch.object(
            elf.ElfFile,
            "_extract_attributes",
            new_callable=lambda: _fake_elffile_extract_attributes,
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        self._elf_files = {
            "fake_elf-2.26": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-2.26")
            ),
            "fake_elf-2.23": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-2.23")
            ),
            "fake_elf-1.1": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-1.1")
            ),
            "fake_elf-static": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-static")
            ),
            "fake_elf-shared-object": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-shared-object")
            ),
            "fake_elf-with-host-libraries": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-with-host-libraries")
            ),
            "fake_elf-bad-ldd": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-bad-ldd")
            ),
            "fake_elf-bad-patchelf": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-bad-patchelf")
            ),
            "fake_elf-with-core-libs": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-with-core-libs")
            ),
            "fake_elf-with-missing-libs": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-with-missing-libs")
            ),
            "fake_elf-with-execstack": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-with-execstack")
            ),
            "fake_elf-with-bad-execstack": elf.ElfFile(
                path=os.path.join(self.root_path, "fake_elf-with-bad-execstack")
            ),
            "libc.so.6": elf.ElfFile(path=os.path.join(self.root_path, "libc.so.6")),
            "libssl.so.1.0.0": elf.ElfFile(
                path=os.path.join(self.root_path, "libssl.so.1.0.0")
            ),
        }

        for elf_file in self._elf_files.values():
            with open(elf_file.path, "wb") as f:
                f.write(b"\x7fELF")
                if elf_file.path.endswith("fake_elf-bad-patchelf"):
                    f.write(b"nointerpreter")

        self.root_libraries = {
            "foo.so.1": os.path.join(self.root_path, "foo.so.1"),
            "moo.so.2": os.path.join(self.root_path, "non-standard", "moo.so.2"),
        }

        barsnap_elf = os.path.join(self.core_base_path, "barsnap.so.2")
        elf_list = [*self.root_libraries.values(), barsnap_elf]

        for root_library in elf_list:
            os.makedirs(os.path.dirname(root_library), exist_ok=True)
            with open(root_library, "wb") as f:
                f.write(b"\x7fELF")


class FakeExtension(fixtures.Fixture):
    """Dynamically generate a new module containing the provided extension"""

    def __init__(self, extension_name, extension_class):
        super().__init__()
        self._import_name = "snapcraft.internal.project_loader._extensions.{}".format(
            extension_name
        )
        self._extension_class = extension_class
        self._extension_name = extension_name

    def _setUp(self):
        extension_module = ModuleType(self._import_name)
        setattr(extension_module, self._extension_class.__name__, self._extension_class)
        sys.modules[self._import_name] = extension_module
        self.addCleanup(self._remove_module)

        json_schema_validator = jsonschema.validate

        def validate(snapcraft, schema, *, format_checker=None):
            try:
                json_schema_validator(snapcraft, schema, format_checker=format_checker)
            except jsonschema.exceptions.ValidationError as error:
                # Ignore validation errors arising from extension usage.
                if (
                    error.instance == self._extension_name
                    and error.path[0] == "apps"
                    and error.path[2] == "extensions"
                ):
                    return None
                raise

        self.useFixture(fixtures.MonkeyPatch("jsonschema.validate", validate))

    def _remove_module(self):
        del sys.modules[self._import_name]


class FakeSnapCommand(fixtures.Fixture):
    def __init__(self):
        self.calls = []
        self.install_success = True
        self.refresh_success = True
        self.download_side_effect = None
        self._email = "-"

    def _setUp(self):
        original_check_call = snapcraft.internal.repo.snaps.check_call
        original_check_output = snapcraft.internal.repo.snaps.check_output

        def side_effect_check_call(cmd, *args, **kwargs):
            return side_effect(original_check_call, cmd, *args, **kwargs)

        def side_effect_check_output(cmd, *args, **kwargs):
            if self._is_snap_command(cmd):
                self.calls.append(cmd)
                return self._fake_snap_command(cmd, *args, **kwargs)
            else:
                return side_effect(original_check_output, cmd, *args, **kwargs)

        def side_effect(original, cmd, *args, **kwargs):
            if self._is_snap_command(cmd):
                self.calls.append(cmd)
                return self._fake_snap_command(cmd, *args, **kwargs)
            else:
                return original(cmd, *args, **kwargs)

        self.useFixture(
            fixtures.MonkeyPatch(
                "snapcraft.internal.repo.snaps.check_call", side_effect_check_call
            )
        )
        self.useFixture(
            fixtures.MonkeyPatch(
                "snapcraft.internal.repo.snaps.check_output", side_effect_check_output
            )
        )

    def login(self, email):
        self._email = email

    def _get_snap_cmd(self, snap_cmd):
        try:
            snap_cmd_index = snap_cmd.index("snap")
        except ValueError:
            return ""

        try:
            return snap_cmd[snap_cmd_index + 1]
        except IndexError:
            return ""

    def _is_snap_command(self, cmd):
        return self._get_snap_cmd(cmd) in ["install", "refresh", "whoami", "download"]

    def _fake_snap_command(self, cmd, *args, **kwargs):
        cmd = self._get_snap_cmd(cmd)
        if cmd == "install" and not self.install_success:
            raise subprocess.CalledProcessError(returncode=1, cmd=cmd)
        elif cmd == "refresh" and not self.refresh_success:
            raise subprocess.CalledProcessError(returncode=1, cmd=cmd)
        elif cmd == "whoami":
            return "email: {}".format(self._email).encode()
        elif (
            cmd == "download"
            and self.download_side_effect is not None
            and not self.download_side_effect.pop(0)
        ):
            raise subprocess.CalledProcessError(returncode=1, cmd=cmd)
        elif cmd == "download":
            return "Downloaded  ".encode()


class FakeSnapcraftctl(fixtures.Fixture):
    def _setUp(self):
        super()._setUp()

        snapcraft_path = get_snapcraft_path()

        tempdir = self.useFixture(fixtures.TempDir()).path
        altered_path = "{}:{}".format(tempdir, os.environ.get("PATH"))
        self.useFixture(fixtures.EnvironmentVariable("PATH", altered_path))

        snapcraftctl_path = os.path.join(tempdir, "snapcraftctl")
        with open(snapcraftctl_path, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                #!/usr/bin/env python3

                # Make sure we can find snapcraft, even if it's not installed
                # (like in CI).
                import sys
                sys.path.append('{snapcraft_path!s}')

                import snapcraft.cli.__main__

                if __name__ == '__main__':
                    snapcraft.cli.__main__.run_snapcraftctl(
                        prog_name='snapcraftctl')
            """.format(
                        snapcraft_path=snapcraft_path
                    )
                )
            )
            f.flush()

        os.chmod(snapcraftctl_path, 0o755)


class FakeMultipass(fixtures.Fixture):
    def _setUp(self):
        super()._setUp()

        tempdir = self.useFixture(fixtures.TempDir()).path
        altered_path = "{}:{}".format(tempdir, os.environ.get("PATH"))
        self.useFixture(fixtures.EnvironmentVariable("PATH", altered_path))

        multipass = os.path.join(tempdir, "multipass")
        os.symlink("/bin/true", multipass)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import types and tell flake8 to ignore the "unused" List.

from typing import Any, Dict, Optional, Tuple

from typing_extensions import Final

from ._extension import Extension


class ExtensionImpl(Extension):
    """Setup a ROS 1 build and runtime environment suitable for a snap."""

    ROS_DISTRO: Final[str] = "noetic"

    @staticmethod
    def get_supported_bases() -> Tuple[str, ...]:
        return ("core20",)

    @staticmethod
    def get_supported_confinement() -> Tuple[str, ...]:
        return ("strict", "devmode")

    @staticmethod
    def is_experimental(base: Optional[str]) -> bool:
        return True

    def __init__(self, *, extension_name: str, yaml_data: Dict[str, Any]) -> None:
        super().__init__(extension_name=extension_name, yaml_data=yaml_data)

        python_paths = [
            f"$SNAP/opt/ros/{self.ROS_DISTRO}/lib/python3.8/site-packages",
            "$SNAP/usr/lib/python3/dist-packages",
            "${PYTHONPATH}",
        ]

        self.root_snippet = {
            "package-repositories": [
                {
                    "type": "apt",
                    "url": "http://packages.ros.org/ros/ubuntu",
                    "components": ["main"],
                    "formats": ["deb"],
                    "key-id": "C1CF6E31E6BADE8868B172B4F42ED6FBAB17C654",
                    "key-server": "keyserver.ubuntu.com",
                    "suites": ["focal"],
                }
            ]
        }

        self.app_snippet = {
            "command-chain": ["snap/command-chain/ros1-launch"],
            "environment": {
                "ROS_DISTRO": self.ROS_DISTRO,
                "PYTHONPATH": ":".join(python_paths),
            },
        }

        self.part_snippet = {"build-environment": [{"ROS_DISTRO": self.ROS_DISTRO}]}

        self.parts = {
            f"ros1-{self.ROS_DISTRO}-extension": {
                "source": "$SNAPCRAFT_EXTENSIONS_DIR/ros1",
                "plugin": "nil",
                "override-build": "install -D -m 0755 launch ${SNAPCRAFT_PART_INSTALL}/snap/command-chain/ros1-launch",
                "build-packages": [f"ros-{self.ROS_DISTRO}-catkin"],
            }
        }



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import Callable, Iterable, Optional, Set, Tuple, Union

from snapcraft import project
from snapcraft.internal import pluginhandler, states, steps

from . import errors


def provides(
    path: str, project: project.Project, parts: Iterable[pluginhandler.PluginHandler]
) -> Set[pluginhandler.PluginHandler]:
    """Determine which part(s) provide(s) the given path.

    :param str path: Path to the directory or file in question.
    :param project.Project project: Project settings.
    :param iterable parts: Iterable of all parts.
    :returns: Set of parts that provide the given path.
    """
    # First of all, ensure the file actually exists before doing any work
    if not os.path.exists(path):
        raise errors.NoSuchFileError(path)

    # Convert file path into absolute path
    absolute_file_path = os.path.abspath(path)

    # Which step are we operating on? We'll know by where the file_path is:
    # the staging area, or the priming area?
    if absolute_file_path.startswith(project.stage_dir):
        step = steps.STAGE
        relative_file_path = os.path.relpath(
            absolute_file_path, start=project.stage_dir
        )
    elif absolute_file_path.startswith(project.prime_dir):
        step = steps.PRIME
        relative_file_path = os.path.relpath(
            absolute_file_path, start=project.prime_dir
        )
    else:
        raise errors.ProvidesInvalidFilePathError(path)

    is_dir = os.path.isdir(absolute_file_path)
    is_file = os.path.isfile(absolute_file_path)

    providing_parts = set()  # type: Set[pluginhandler.PluginHandler]
    for part, state in _part_states_for_step(step, parts):
        if is_dir and relative_file_path in state.directories:
            providing_parts.add(part)
        elif is_file and relative_file_path in state.files:
            providing_parts.add(part)

    if not providing_parts:
        raise errors.UntrackedFileError(path)

    return providing_parts


# Implemented as a generator since loading up the state could be heavy
def _part_states_for_step(
    step: steps.Step, parts: Iterable[pluginhandler.PluginHandler]
) -> Iterable[
    Tuple[pluginhandler.PluginHandler, Union[states.StageState, states.PrimeState]]
]:
    state_getter: Optional[
        Callable[
            [pluginhandler.PluginHandler], Union[states.StageState, states.PrimeState]
        ]
    ] = None

    if step == steps.STAGE:
        state_getter = pluginhandler.PluginHandler.get_stage_state
    elif step == steps.PRIME:
        state_getter = pluginhandler.PluginHandler.get_prime_state
    else:
        # This is programmer error
        raise RuntimeError("Only the stage or prime step is supported!")

    for part in parts:
        state = state_getter(part)
        if state:
            yield (part, state)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import os
from collections import OrderedDict
from typing import List, Optional, Tuple

from tabulate import tabulate
from typing_extensions import Final

from snapcraft.storeapi.v2.channel_map import (
    ChannelMap,
    MappedChannel,
    Revision,
    SnapChannel,
)


class _HINTS:
    CLOSED: Final[str] = "-"
    FOLLOWING: Final[str] = "↑"
    NO_PROGRESS: Final[str] = "-"
    PROGRESSING_TO: Final[str] = "→"
    UNKNOWN: Final[str] = "?"


def _get_channel_order(
    snap_channels, tracks: Optional[List[str]] = None
) -> OrderedDict:
    channel_order: OrderedDict = OrderedDict()

    if tracks is not None:
        snap_channels = [s for s in snap_channels if s.track in tracks]

    for snap_channel in snap_channels:
        if snap_channel.track not in channel_order:
            channel_order[snap_channel.track] = list()
        if snap_channel.fallback is None:
            channel_order[snap_channel.track].append(snap_channel.name)
        else:
            try:
                channel_order[snap_channel.track].insert(
                    channel_order[snap_channel.track].index(snap_channel.fallback) + 1,
                    snap_channel.name,
                )
            except ValueError:
                channel_order[snap_channel.track].append(snap_channel.name)

    return channel_order


def _get_channel_line(
    *,
    mapped_channel: Optional[MappedChannel],
    revision: Optional[Revision],
    channel_info: SnapChannel,
    hint: str,
    progress_string: str,
) -> List[str]:
    version_string = hint
    revision_string = hint
    expiration_date_string = ""
    channel_string = channel_info.risk

    if revision is not None:
        version_string = revision.version
        revision_string = f"{revision.revision}"

    if mapped_channel is not None:
        if channel_info.branch is None and mapped_channel.progressive.percentage:
            channel_string = ""
        elif channel_info.branch is not None:
            channel_string = f"{channel_info.risk}/{channel_info.branch}"
            if mapped_channel.expiration_date is not None:
                expiration_date_string = mapped_channel.expiration_date

    if os.getenv("SNAPCRAFT_EXPERIMENTAL_PROGRESSIVE_RELEASES"):
        return [
            channel_string,
            version_string,
            revision_string,
            progress_string,
            expiration_date_string,
        ]
    else:
        return [channel_string, version_string, revision_string, expiration_date_string]


def _get_channel_lines_for_channel(  # noqa: C901
    snap_channel_map: ChannelMap,
    channel_name: str,
    architecture: str,
    current_tick: str,
) -> Tuple[str, List[List[str]]]:
    channel_lines: List[List[str]] = list()

    channel_info = snap_channel_map.get_channel_info(channel_name)

    try:
        progressive_mapped_channel: Optional[MappedChannel] = (
            snap_channel_map.get_mapped_channel(
                channel_name=channel_name, architecture=architecture, progressive=True
            )
        )
    except ValueError:
        progressive_mapped_channel = None

    if progressive_mapped_channel is not None:
        progressive_revision = snap_channel_map.get_revision(
            progressive_mapped_channel.revision
        )

        if progressive_mapped_channel.progressive.percentage is None:
            raise RuntimeError("Unexpected null progressive percentage")
        else:
            percentage = progressive_mapped_channel.progressive.percentage

        if progressive_mapped_channel.progressive.current_percentage is None:
            current_percentage_fmt = _HINTS.UNKNOWN
            remaining_percentage_fmt = _HINTS.UNKNOWN
        else:
            current_percentage = (
                progressive_mapped_channel.progressive.current_percentage
            )
            current_percentage_fmt = f"{current_percentage:.0f}"
            remaining_percentage_fmt = f"{100 - current_percentage:.0f}"

        progressive_mapped_channel_line = _get_channel_line(
            mapped_channel=progressive_mapped_channel,
            revision=progressive_revision,
            channel_info=channel_info,
            hint=current_tick,
            progress_string=f"{current_percentage_fmt}{_HINTS.PROGRESSING_TO}{percentage:.0f}%",
        )
        # Setup progress for the actually released revision, this needs to be
        # calculated. But only show it if the channel is open.
        progress_string = (
            f"{remaining_percentage_fmt}{_HINTS.PROGRESSING_TO}{100 - percentage:.0f}%"
        )
    else:
        progress_string = _HINTS.NO_PROGRESS

    try:
        mapped_channel: Optional[MappedChannel] = snap_channel_map.get_mapped_channel(
            channel_name=channel_name, architecture=architecture, progressive=False
        )
    except ValueError:
        mapped_channel = None

    next_tick = current_tick
    if mapped_channel is not None:
        revision = snap_channel_map.get_revision(mapped_channel.revision)
        channel_lines.append(
            _get_channel_line(
                mapped_channel=mapped_channel,
                revision=revision,
                channel_info=channel_info,
                hint=current_tick,
                progress_string=progress_string,
            )
        )
        if channel_info.branch is None:
            next_tick = _HINTS.FOLLOWING
    # Show an empty entry if there is no specific channel information, but
    # only for <track>/<risks> (ignoring /<branch>).
    elif channel_info.branch is None:
        channel_lines.append(
            _get_channel_line(
                mapped_channel=None,
                revision=None,
                channel_info=channel_info,
                hint=current_tick,
                progress_string=_HINTS.NO_PROGRESS
                if current_tick == _HINTS.CLOSED
                else progress_string,
            )
        )

    if (
        os.getenv("SNAPCRAFT_EXPERIMENTAL_PROGRESSIVE_RELEASES")
        and progressive_mapped_channel is not None
    ):
        channel_lines.append(progressive_mapped_channel_line)
        if channel_info.branch is None:
            next_tick = _HINTS.FOLLOWING

    return next_tick, channel_lines


def _has_channels_for_architecture(
    snap_channel_map, architecture: str, channels: List[str]
) -> bool:
    progressive = (False, True)
    # channel_query = (channel_name, progressive)
    for channel_query in itertools.product(channels, progressive):
        try:
            snap_channel_map.get_mapped_channel(
                channel_name=channel_query[0],
                architecture=architecture,
                progressive=channel_query[1],
            )
            found_architecture = True
            break
        except ValueError:
            continue
    else:
        found_architecture = False

    return found_architecture


def get_tabulated_channel_map(
    snap_channel_map, *, architectures: List[str], tracks: Optional[List[str]] = None
):
    channel_order = _get_channel_order(snap_channel_map.snap.channels, tracks)

    channel_lines = list()
    for track_name in channel_order:
        track_mentioned = False
        for architecture in sorted(architectures):
            if not _has_channels_for_architecture(
                snap_channel_map, architecture, channel_order[track_name]
            ):
                continue
            architecture_mentioned = False
            next_tick = _HINTS.CLOSED
            for channel_name in channel_order[track_name]:
                if not track_mentioned:
                    track_mentioned = True
                    track_string = track_name
                else:
                    track_string = ""

                if not architecture_mentioned:
                    architecture_mentioned = True
                    architecture_string = architecture
                else:
                    architecture_string = ""

                next_tick, parsed_channels = _get_channel_lines_for_channel(
                    snap_channel_map, channel_name, architecture, next_tick
                )
                for channel_line in parsed_channels:
                    channel_lines.append(
                        [track_string, architecture_string] + channel_line
                    )
                    track_string = ""
                    architecture_string = ""

    headers = ["Track", "Arch", "Channel", "Version", "Revision"]
    if os.getenv("SNAPCRAFT_EXPERIMENTAL_PROGRESSIVE_RELEASES"):
        headers.append("Progress")
        # Item 6 is expiration_date when progressive releases are enabled.
        expires_column = 6
    else:
        expires_column = 5

    if any(line[expires_column] != "" for line in channel_lines):
        headers.append("Expires at")
    else:
        headers.append("")

    return tabulate(channel_lines, numalign="left", headers=headers, tablefmt="plain")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from collections import ChainMap
from os import path
from typing import Set  # noqa: F401
from typing import List

import snapcraft
from snapcraft.internal import elf, pluginhandler, repo
from snapcraft.internal.pluginhandler._part_environment import (
    get_snapcraft_global_environment,
    get_snapcraft_part_directory_environment,
)

from . import errors, grammar_processing
from ._env import build_env, build_env_for_stage, runtime_env

logger = logging.getLogger(__name__)


class PartsConfig:
    def __init__(self, *, parts, project, validator):
        self._soname_cache = elf.SonameCache()
        self._parts_data = parts.get("parts", {})
        self._snap_type = parts.get("type", "app")
        self._project = project
        self._validator = validator

        self.all_parts = []
        self._part_names = []
        self.after_requests = {}

        self._process_parts()

    @property
    def part_names(self):
        return self._part_names

    def _process_parts(self):
        for part_name in self._parts_data:
            self._part_names.append(part_name)
            properties = self._parts_data[part_name] or {}

            plugin_name = properties.get("plugin")

            if "after" in properties:
                self.after_requests[part_name] = properties.pop("after")

            if "filesets" in properties:
                del properties["filesets"]

            self.load_part(part_name, plugin_name, properties)

        self._compute_dependencies()
        self.all_parts = self._sort_parts()

    def _compute_dependencies(self):
        """Gather the lists of dependencies and adds to all_parts."""

        for part in self.all_parts:
            dep_names = self.after_requests.get(part.name, [])
            for dep_name in dep_names:
                dep = self.get_part(dep_name)
                if not dep:
                    raise errors.SnapcraftAfterPartMissingError(part.name, dep_name)

                part.deps.append(dep)

    def _sort_parts(self):
        """Performs an inneficient but easy to follow sorting of parts."""
        sorted_parts = []

        # We want to process parts in a consistent order between runs. The
        # simplest way to do this is to sort them by name.
        self.all_parts = sorted(
            self.all_parts, key=lambda part: part.name, reverse=True
        )

        while self.all_parts:
            top_part = None
            for part in self.all_parts:
                mentioned = False
                for other in self.all_parts:
                    if part in other.deps:
                        mentioned = True
                        break
                if not mentioned:
                    top_part = part
                    break
            if not top_part:
                raise errors.SnapcraftLogicError(
                    "circular dependency chain found in parts definition"
                )
            sorted_parts = [top_part] + sorted_parts
            self.all_parts.remove(top_part)

        return sorted_parts

    def get_dependencies(
        self, part_name: str, *, recursive: bool = False
    ) -> Set[pluginhandler.PluginHandler]:
        """Returns a set of all the parts upon which part_name depends."""

        dependency_names = set(self.after_requests.get(part_name, []))
        dependencies = {p for p in self.all_parts if p.name in dependency_names}

        if recursive:
            # No need to worry about infinite recursion due to circular
            # dependencies since the YAML validation won't allow it.
            for dependency_name in dependency_names:
                dependencies |= self.get_dependencies(
                    dependency_name, recursive=recursive
                )

        return dependencies

    def get_reverse_dependencies(
        self, part_name: str, *, recursive: bool = False
    ) -> Set[pluginhandler.PluginHandler]:
        """Returns a set of all the parts that depend upon part_name."""

        reverse_dependency_names = set()
        for part, dependencies in self.after_requests.items():
            if part_name in dependencies:
                reverse_dependency_names.add(part)

        reverse_dependencies = {
            p for p in self.all_parts if p.name in reverse_dependency_names
        }

        if recursive:
            # No need to worry about infinite recursion due to circular
            # dependencies since the YAML validation won't allow it.
            for reverse_dependency_name in reverse_dependency_names:
                reverse_dependencies |= self.get_reverse_dependencies(
                    reverse_dependency_name, recursive=recursive
                )

        return reverse_dependencies

    def get_part(self, part_name):
        for part in self.all_parts:
            if part.name == part_name:
                return part

        return None

    def clean_part(self, part_name, staged_state, primed_state, step):
        part = self.get_part(part_name)
        part.clean(staged_state, primed_state, step)

    def validate(self, part_names):
        for part_name in part_names:
            if part_name not in self._part_names:
                raise snapcraft.internal.errors.SnapcraftEnvironmentError(
                    "The part named {!r} is not defined in "
                    "{!r}".format(
                        part_name, self._project.info.snapcraft_yaml_file_path
                    )
                )

    def load_part(self, part_name, plugin_name, part_properties):
        plugin = pluginhandler.load_plugin(
            plugin_name=plugin_name,
            part_name=part_name,
            properties=part_properties,
            project=self._project,
            part_schema=self._validator.part_schema,
            definitions_schema=self._validator.definitions_schema,
        )

        logger.debug(
            "Setting up part {!r} with plugin {!r} and "
            "properties {!r}.".format(part_name, plugin_name, part_properties)
        )

        stage_packages_repo = repo.Repo

        grammar_processor = grammar_processing.PartGrammarProcessor(
            plugin=plugin,
            properties=part_properties,
            project=self._project,
            repo=stage_packages_repo,
        )

        part = pluginhandler.PluginHandler(
            plugin=plugin,
            part_properties=part_properties,
            project=self._project,
            part_schema=self._validator.part_schema,
            definitions_schema=self._validator.definitions_schema,
            stage_packages_repo=stage_packages_repo,
            grammar_processor=grammar_processor,
            snap_base_path=path.join("/", "snap", self._project.info.name, "current"),
            soname_cache=self._soname_cache,
        )

        self.all_parts.append(part)

        return part

    def build_env_for_part(self, part, root_part=True) -> List[str]:
        """Return a build env of all the part's dependencies."""

        env = []  # type: List[str]
        stagedir = self._project.stage_dir

        if root_part:
            # this has to come before any {}/usr/bin
            env += part.env(part.part_install_dir)
            env += runtime_env(part.part_install_dir, self._project.arch_triplet)
            env += runtime_env(stagedir, self._project.arch_triplet)
            env += build_env(
                part.part_install_dir,
                self._project.info.name,
                self._project.arch_triplet,
            )
            env += build_env_for_stage(
                stagedir, self._project.info.name, self._project.arch_triplet
            )

            global_env = get_snapcraft_global_environment(self._project)
            part_env = get_snapcraft_part_directory_environment(part)

            for variable, value in ChainMap(part_env, global_env).items():
                env.append('{}="{}"'.format(variable, value))

            # Finally, add the declared environment from the part.
            # This is done only for the "root" part.
            for be in part.build_environment:
                env.extend([f'{k}="{v}"' for k, v in be.items()])
        else:
            env += part.env(stagedir)
            env += runtime_env(stagedir, self._project.arch_triplet)

        for dep_part in part.deps:
            env += dep_part.env(stagedir)
            env += self.build_env_for_part(dep_part, root_part=False)

        # LP: #1767625
        # Remove duplicates from using the same plugin in dependent parts.
        seen = set()  # type: Set[str]
        deduped_env = list()  # type: List[str]
        for e in env:
            if e not in seen:
                deduped_env.append(e)
                seen.add(e)

        return deduped_env



#!/usr/bin/env python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os
import subprocess
import tempfile

ACTIVE_DISTROS = ("xenial", "artful", "bionic")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("day", help="The day of the results, with format yyyymmdd")
    args = parser.parse_args()
    install_autopkgtest_results_formatter()
    with tempfile.TemporaryDirectory(dir=os.environ.get("HOME")) as temp_dir:
        clone_results_repo(temp_dir)
        format_results(temp_dir, ACTIVE_DISTROS, args.day)
        commit_and_push(temp_dir, args.day)


def install_autopkgtest_results_formatter():
    subprocess.check_call(
        ["sudo", "snap", "install", "autopkgtest-results-formatter", "--edge"]
    )


def clone_results_repo(dest_dir):
    subprocess.check_call(
        ["git", "clone", "https://github.com/elopio/autopkgtest-results.git", dest_dir]
    )


def format_results(dest_dir, distros, day):
    subprocess.check_call(
        [
            "/snap/bin/autopkgtest-results-formatter",
            "--destination",
            dest_dir,
            "--distros",
            *distros,
            "--day",
            day,
        ]
    )


def commit_and_push(repo_dir, day):
    subprocess.check_call(
        ["git", "config", "--global", "user.email", "u1test+m-o@canonical.com"]
    )
    subprocess.check_call(["git", "config", "--global", "user.name", "snappy-m-o"])

    subprocess.check_call(["git", "-C", repo_dir, "add", "--all"])
    subprocess.check_call(
        [
            "git",
            "-C",
            repo_dir,
            "commit",
            "--message",
            "Add the results for {}".format(day),
        ]
    )

    subprocess.check_call(
        [
            "git",
            "-C",
            repo_dir,
            "push",
            "https://{GH_TOKEN}@github.com/elopio/autopkgtest-results.git".format(
                GH_TOKEN=os.environ.get("GH_TOKEN_PPA_AUTOPKGTEST_RESULTS")
            ),
        ]
    )


if __name__ == "__main__":
    main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import textwrap

from testtools.matchers import Equals

from snapcraft.internal import steps
from snapcraft.internal.project_loader import inspection

from .. import ProjectLoaderBaseTest


class LifecycleStatusTest(ProjectLoaderBaseTest):
    def setUp(self):
        super().setUp()

        for path in ("src1", "src2"):
            os.mkdir(path)

        self.config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                name: my-snap-name
                base: core18
                version: '1.0'
                summary: summary
                description: description

                grade: devel
                confinement: devmode

                parts:
                    part1:
                        plugin: nil
                        source: src1/
                    part2:
                        plugin: nil
                        source: src2/
                        after: [part1]
                """
            )
        )

        for part in self.config.all_parts:
            part.makedirs()

        self.part1 = self.config.parts.get_part("part1")
        self.part2 = self.config.parts.get_part("part2")

    def test_lifecycle_status(self):
        # Stage part1, since it's a dependency
        self.part1.mark_pull_done()
        self.part1.mark_build_done()
        self.part1.mark_stage_done(set(), set())

        # Pull part2
        self.part2.mark_pull_done()

        self.assertThat(
            inspection.lifecycle_status(self.config),
            Equals(
                [
                    {
                        "part": "part1",
                        "pull": "complete",
                        "build": "complete",
                        "stage": "complete",
                        "prime": None,
                    },
                    {
                        "part": "part2",
                        "pull": "complete",
                        "build": None,
                        "stage": None,
                        "prime": None,
                    },
                ]
            ),
        )

        # Now prime them both
        self.part1.mark_prime_done(set(), set(), set(), set())
        self.part2.mark_build_done()
        self.part2.mark_stage_done(set(), set())
        self.part2.mark_prime_done(set(), set(), set(), set())

        self.assertThat(
            inspection.lifecycle_status(self.config),
            Equals(
                [
                    {
                        "part": "part1",
                        "pull": "complete",
                        "build": "complete",
                        "stage": "complete",
                        "prime": "complete",
                    },
                    {
                        "part": "part2",
                        "pull": "complete",
                        "build": "complete",
                        "stage": "complete",
                        "prime": "complete",
                    },
                ]
            ),
        )

        # Change the source of part2, which should make its pull step outdated
        open(os.path.join("src2", "file"), "w").close()
        self.assertThat(
            inspection.lifecycle_status(self.config),
            Equals(
                [
                    {
                        "part": "part1",
                        "pull": "complete",
                        "build": "complete",
                        "stage": "complete",
                        "prime": "complete",
                    },
                    {
                        "part": "part2",
                        "pull": "outdated (source changed)",
                        "build": "complete",
                        "stage": "complete",
                        "prime": "complete",
                    },
                ]
            ),
        )

        # Now clean the prime step of part1 and verify that it effects part2
        self.part1.mark_cleaned(steps.PRIME)
        self.assertThat(
            inspection.lifecycle_status(self.config),
            Equals(
                [
                    {
                        "part": "part1",
                        "pull": "complete",
                        "build": "complete",
                        "stage": "complete",
                        "prime": None,
                    },
                    {
                        "part": "part2",
                        "pull": "outdated (source changed)",
                        "build": "complete",
                        "stage": "complete",
                        "prime": "dirty ('part1' changed)",
                    },
                ]
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from testtools.matchers import Equals

from . import CommandBaseTestCase


class VersionCommandTestCase(CommandBaseTestCase):
    def test_has_version(self):
        result = self.run_command(["--version"])
        self.assertThat(result.exit_code, Equals(0))

    def test_has_version_without_hyphens(self):
        result = self.run_command(["version"])
        self.assertThat(result.exit_code, Equals(0))

    def test_method_return_same_value(self):
        result1 = self.run_command(["version"])
        result2 = self.run_command(["--version"])
        self.assertEqual(result1.output, result2.output)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testtools.matchers import Equals

from snapcraft.internal import errors, os_release
from tests import unit


class OsReleaseTestCase(unit.TestCase):
    def _write_os_release(self, contents):
        path = "os-release"
        with open(path, "w") as f:
            f.write(contents)
        return path

    def test_blank_lines(self):
        release = os_release.OsRelease(
            os_release_file=self._write_os_release(
                dedent(
                    """\
                NAME="Arch Linux"

                PRETTY_NAME="Arch Linux"
                ID=arch
                ID_LIKE=archlinux
                VERSION_ID="foo"
                VERSION_CODENAME="bar"

            """
                )
            )
        )

        self.assertThat(release.id(), Equals("arch"))
        self.assertThat(release.name(), Equals("Arch Linux"))
        self.assertThat(release.version_id(), Equals("foo"))
        self.assertThat(release.version_codename(), Equals("bar"))

    def test_no_id(self):
        release = os_release.OsRelease(
            os_release_file=self._write_os_release(
                dedent(
                    """\
                NAME="Arch Linux"
                PRETTY_NAME="Arch Linux"
                ID_LIKE=archlinux
                VERSION_ID="foo"
                VERSION_CODENAME="bar"
            """
                )
            )
        )

        self.assertRaises(errors.OsReleaseIdError, release.id)

    def test_no_name(self):
        release = os_release.OsRelease(
            os_release_file=self._write_os_release(
                dedent(
                    """\
                ID=arch
                PRETTY_NAME="Arch Linux"
                ID_LIKE=archlinux
                VERSION_ID="foo"
                VERSION_CODENAME="bar"
            """
                )
            )
        )

        self.assertRaises(errors.OsReleaseNameError, release.name)

    def test_no_version_id(self):
        release = os_release.OsRelease(
            os_release_file=self._write_os_release(
                dedent(
                    """\
                NAME="Arch Linux"
                ID=arch
                PRETTY_NAME="Arch Linux"
                ID_LIKE=archlinux
                VERSION_CODENAME="bar"
            """
                )
            )
        )

        self.assertRaises(errors.OsReleaseVersionIdError, release.version_id)

    def test_no_version_codename(self):
        """Test that version codename can also come from VERSION_ID"""
        release = os_release.OsRelease(
            os_release_file=self._write_os_release(
                dedent(
                    """\
                NAME="Ubuntu"
                VERSION="14.04.5 LTS, Trusty Tahr"
                ID=ubuntu
                ID_LIKE=debian
                PRETTY_NAME="Ubuntu 14.04.5 LTS"
                VERSION_ID="14.04"
            """
                )
            )
        )

        self.assertThat(release.version_codename(), Equals("trusty"))

    def test_no_version_codename_or_version_id(self):
        release = os_release.OsRelease(
            os_release_file=self._write_os_release(
                dedent(
                    """\
                NAME="Ubuntu"
                ID=ubuntu
                ID_LIKE=debian
                PRETTY_NAME="Ubuntu 16.04.3 LTS"
            """
                )
            )
        )

        self.assertRaises(errors.OsReleaseCodenameError, release.version_codename)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import time
from typing import List

import click
from xdg import BaseDirectory

from snapcraft.formatting_utils import humanize_list
from snapcraft.internal.remote_build import LaunchpadClient, WorkTree, errors
from snapcraft.project import Project

from . import echo
from ._options import PromptOption, get_project

_SUPPORTED_ARCHS = ["amd64", "arm64", "armhf", "i386", "ppc64el", "s390x"]


@click.group()
def remotecli():
    """Remote build commands"""


@remotecli.command("remote-build")
@click.option("--recover", is_flag=True, help="Recover interrupted build.")
@click.option("--status", is_flag=True, help="Display remote build status.")
@click.option(
    "--build-on",
    metavar="<arch-list>",
    type=str,
    nargs=1,
    required=False,
    help="Set architectures to build on.",
)
@click.option(
    "--launchpad-accept-public-upload",
    is_flag=True,
    prompt=(
        "All data sent to remote builders will be publicly available. "
        "Are you sure you want to continue?"
    ),
    help="Acknowledge that uploaded code will be publicly available.",
    cls=PromptOption,
)
@click.option(
    "--launchpad-timeout",
    metavar="<seconds>",
    type=int,
    nargs=1,
    required=False,
    help=(
        "Time to wait for Launchpad to build before exiting ('0' disables timeout)."
        "Note that the build will continue on Launchpad and can be resumed later."
    ),
    default=0,
)
@click.option(
    "--package-all-sources",
    is_flag=True,
    help="Package all sources to send to remote builder, not just local sources.",
)
def remote_build(
    recover: bool,
    status: bool,
    build_on: str,
    launchpad_accept_public_upload: bool,
    launchpad_timeout: int,
    package_all_sources: bool,
    echoer=echo,
) -> None:
    """Dispatch a snap for remote build.

    Command remote-build sends the current project to be built remotely. After the build
    is complete, packages for each architecture are retrieved and will be available in
    the local filesystem.

    If not specified in the snapcraft.yaml file, the list of architectures to build
    can be set using the --build-on option. If both are specified, an error will occur.

    Interrupted remote builds can be resumed using the --recover option, followed by
    the build number informed when the remote build was originally dispatched. The
    current state of the remote build for each architecture can be checked using the
    --status option.

    \b
    Examples:
        snapcraft remote-build
        snapcraft remote-build --build-on=amd64
        snapcraft remote-build --build-on=amd64,arm64,armhf,i386,ppc64el,s390x
        snapcraft remote-build --recover
        snapcraft remote-build --status
    """
    if os.getenv("SUDO_USER") and os.geteuid() == 0:
        echo.warning(
            "Running with 'sudo' may cause permission errors and is discouraged."
        )

    if not launchpad_accept_public_upload:
        raise errors.AcceptPublicUploadError()

    echo.warning(
        "snapcraft remote-build is experimental and is subject to change - use with caution."
    )

    project = get_project()

    # TODO: use project.is_legacy() when available.
    try:
        project._get_build_base()
    except RuntimeError:
        raise errors.BaseRequiredError()

    # Use a hash of current working directory to distinguish between other
    # potential project builds occurring in parallel elsewhere.
    project_hash = project._get_project_directory_hash()
    build_id = f"snapcraft-{project.info.name}-{project_hash}"
    architectures = _determine_architectures(project, build_on)

    # Calculate timeout timestamp, if specified.
    if launchpad_timeout > 0:
        deadline = int(time.time()) + launchpad_timeout
    else:
        deadline = 0

    lp = LaunchpadClient(
        project=project,
        build_id=build_id,
        architectures=architectures,
        deadline=deadline,
    )

    if status:
        _print_status(lp)
        return

    if lp.has_outstanding_build():
        echo.info("Found previously started build.")
        _print_status(lp)

        # If recovery specified, monitor build and exit.
        if recover or echo.confirm("Do you wish to recover this build?", default=True):
            _monitor_build(lp)
            return

        # Otherwise clean running build before we start a new one.
        _clean_build(lp)

    _start_build(
        lp=lp,
        project=project,
        build_id=build_id,
        package_all_sources=package_all_sources,
    )

    _monitor_build(lp)


def _clean_build(lp: LaunchpadClient):
    echo.info("Cleaning existing builds and artifacts...")
    lp.cleanup()
    echo.info("Done.")


def _print_status(lp: LaunchpadClient):
    if lp.has_outstanding_build():
        build_status = lp.get_build_status()
        for arch, status in build_status.items():
            echo.info(f"Build status for arch {arch}: {status}")
    else:
        echo.info("No build found.")


def _start_build(
    *, lp: LaunchpadClient, project: Project, build_id: str, package_all_sources: bool
) -> None:
    # Pull/update sources for project.
    worktree_dir = BaseDirectory.save_data_path("snapcraft", "remote-build", build_id)
    wt = WorkTree(worktree_dir, project, package_all_sources=package_all_sources)
    repo_dir = wt.prepare_repository()
    lp.push_source_tree(repo_dir)

    # Start building.
    lp.start_build()
    echo.info("If interrupted, resume with: 'snapcraft remote-build --recover'")


def _monitor_build(lp: LaunchpadClient) -> None:
    target_list = humanize_list(lp.architectures, "and", "{}")
    echo.info(
        f"Building snap package for {target_list}. This may take some time to finish."
    )

    lp.monitor_build()

    echo.info("Build complete.")
    lp.cleanup()


def _check_supported_architectures(archs: List[str]) -> None:
    unsupported_archs = []
    for item in archs:
        if item not in _SUPPORTED_ARCHS:
            unsupported_archs.append(item)
    if unsupported_archs:
        raise errors.UnsupportedArchitectureError(archs=unsupported_archs)


def _get_project_architectures(project) -> List[str]:
    archs = []
    if project.info.architectures:
        for item in project.info.architectures:
            if "build-on" in item:
                new_arch = item["build-on"]
                if isinstance(new_arch, list):
                    # FIXME: LP builders don't recognize a list of architectures
                    archs.extend(new_arch)
                else:
                    archs.append(new_arch)

    return archs


def _determine_architectures(project: Project, user_specified_arch: str):
    # If build architectures not set in snapcraft.yaml, let the user override
    # Launchpad defaults using --build-on.
    project_architectures = _get_project_architectures(project)
    if project_architectures and user_specified_arch:
        raise click.BadOptionUsage(
            "--build-on",
            "Cannot use --build-on, architecture list is already set in snapcraft.yaml.",
        )

    if project_architectures:
        archs = project_architectures
    elif user_specified_arch:
        archs = user_specified_arch.split(",")
    else:
        # Default to typical snapcraft behavior (build for host).
        archs = [project.deb_arch]

    # Sanity check for build architectures
    _check_supported_architectures(archs)

    return archs



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import pathlib
from textwrap import dedent

import pytest

import snapcraft.internal.errors
from snapcraft.project import Project
from snapcraft.project._sanity_checks import conduct_project_sanity_check


@pytest.fixture
def snapcraft_yaml_path(tmp_work_path, request):
    """Generic snapcraft application project with a nil part and no apps."""
    snapcraft_yaml_path = tmp_work_path / "snap/snapcraft.yaml"
    snapcraft_yaml_path.parent.mkdir(parents=True)
    with snapcraft_yaml_path.open("w") as snapcraft_file:
        print(
            dedent(
                """\
            name: project-name
            base: core18
            version: "1.0"
            summary: sanity checks
            description: sanity checks
            grade: stable
            confinement: strict

            parts:
              nil:
                plugin: nil
            """
            ),
            file=snapcraft_file,
        )

    return snapcraft_yaml_path.relative_to(tmp_work_path)


@pytest.fixture(params=[True, False])
def project(snapcraft_yaml_path, request):
    """Return a project in host and managed-host modes."""
    snapcraft_project = Project(
        is_managed_host=request.param,
        snapcraft_yaml_file_path=snapcraft_yaml_path.as_posix(),
    )
    return snapcraft_project


@pytest.fixture
def caplog_warning(caplog):
    """Return caplog set to logging.WARNING."""
    caplog.set_level(logging.WARNING)
    return caplog


def test_no_snap_dir(caplog_warning, project):
    conduct_project_sanity_check(project)
    assert len(caplog_warning.records) == 0


def test_icon(tmp_work_path):
    snapcraft_yaml_path = tmp_work_path / "snap/snapcraft.yaml"
    snapcraft_yaml_path.parent.mkdir(parents=True)
    with snapcraft_yaml_path.open("w") as snapcraft_file:
        print(
            dedent(
                """\
            name: project-name
            base: core18
            version: "1.0"
            summary: sanity checks
            description: sanity checks
            grade: stable
            confinement: strict
            icon: foo.png

            parts:
              nil:
                plugin: nil
            """
            ),
            file=snapcraft_file,
        )

    project = Project(
        is_managed_host=False, snapcraft_yaml_file_path=snapcraft_yaml_path.as_posix(),
    )

    # Test without icon raises error
    with pytest.raises(snapcraft.internal.errors.SnapcraftEnvironmentError) as exc_info:
        conduct_project_sanity_check(project)

    assert exc_info.value.get_brief() == "Specified icon 'foo.png' does not exist."

    # Test with icon passes.
    (tmp_work_path / "foo.png").touch()
    conduct_project_sanity_check(project)


def test_accepted_artifacts(caplog_warning, project):
    assets_dir = pathlib.Path("snap")

    file_assets = [
        ".snapcraft/state",
        "gui/icon.png",
        "gui/other-icon.png",
        "plugins/plugin1.py",
        "plugins/data-file",
        "hooks/configure",
        "hooks/new-hook",
        "keys/key1.asc",
        "keys/key2.asc",
        "local/file",
        "local/dir/file",
    ]

    for file_asset in file_assets:
        asset_path = assets_dir / file_asset
        asset_path.parent.mkdir(parents=True, exist_ok=True)
        asset_path.touch()

    conduct_project_sanity_check(project)

    assert len(caplog_warning.records) == 0


def test_unexpected_things(caplog_warning, project):
    assets_dir = pathlib.Path("snap")

    file_assets = [
        "dir1/foo",
        "dir1/keys/key1.asc",
        "dir1/keys/key2.asc",
        "dir1/local/dir/file",
        "dir1/local/file",
        "dir1/plugins/data-file",
        "dir1/plugins/plugin1.py",
        "dir2/foo",
        "dir2/hooks/configure",
        "dir2/hooks/new-hook",
        "gui/icon.jpg",
    ]

    for file_asset in file_assets:
        asset_path = assets_dir / file_asset
        asset_path.parent.mkdir(parents=True, exist_ok=True)
        asset_path.touch()

    conduct_project_sanity_check(project)
    assert caplog_warning.records[0].message == (
        "The 'snap' directory is meant specifically for snapcraft, but it "
        "contains the following non-snapcraft-related paths, which is "
        "unsupported and will cause unexpected behavior:\n"
        "- dir1\n"
        "- dir1/foo\n"
        "- dir1/keys\n"
        "- dir1/keys/key1.asc\n"
        "- dir1/keys/key2.asc\n"
        "- dir1/local\n"
        "- dir1/local/dir\n"
        "- dir1/local/dir/file\n"
        "- dir1/local/file\n"
        "- dir1/plugins\n"
        "- dir1/plugins/data-file\n"
        "- dir1/plugins/plugin1.py\n"
        "- dir2\n"
        "- dir2/foo\n"
        "- dir2/hooks\n"
        "- dir2/hooks/configure\n"
        "- dir2/hooks/new-hook\n"
        "- gui/icon.jpg\n"
        "\n"
        "If you must store these files within the 'snap' directory, move them "
        "to 'snap/local', which is ignored by snapcraft."
    )



import os
from urllib.parse import urljoin

import requests

from ._requests import Requests
from . import constants


class UpDownClient(Requests):
    """The Up/Down server provide upload/download snap capabilities."""

    def __init__(self, client) -> None:
        self._client = client
        self._root_url = os.getenv("STORE_UPLOAD_URL", constants.STORE_UPLOAD_URL)

    def _request(self, method, urlpath, **kwargs) -> requests.Response:
        url = urljoin(self._root_url, urlpath)
        return self._client.request(method, url, **kwargs)

    def upload(self, monitor):
        return self.post(
            "/unscanned-upload/",
            data=monitor,
            headers={
                "Content-Type": monitor.content_type,
                "Accept": "application/json",
            },
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from snapcraft.internal import sources


class TestUri:

    scenarios = [
        ("tar.gz", dict(result="tar", source="https://golang.tar.gz")),
        ("tar.gz", dict(result="tar", source="https://golang.tar.xz")),
        ("tar.bz2", dict(result="tar", source="https://golang.tar.bz2")),
        ("tgz", dict(result="tar", source="https://golang.tgz")),
        ("tar", dict(result="tar", source="https://golang.tar")),
        ("git:", dict(result="git", source="git://github.com:snapcore/snapcraft.git")),
        ("git@", dict(result="git", source="git@github.com:snapcore/snapcraft.git")),
        (
            ".git",
            dict(result="git", source="https://github.com:snapcore/snapcraft.git"),
        ),
        ("lp:", dict(result="bzr", source="lp:snapcraft_test_source")),
        ("bzr:", dict(result="bzr", source="bzr:dummy_source")),
        ("svn:", dict(result="subversion", source="svn://sylpheed.jp/sylpheed/trunk")),
    ]

    def test(self, source, result):
        assert sources._get_source_type_from_uri(source) == result


class TestSourceWithBranchErrors:

    scenarios = [
        (
            "bzr with source branch",
            {
                "source_type": "bzr",
                "source_branch": "test_branch",
                "source_tag": None,
                "source_commit": None,
                "error": "source-branch",
            },
        ),
        (
            "tar with source branch",
            {
                "source_type": "tar",
                "source_branch": "test_branch",
                "source_tag": None,
                "source_commit": None,
                "error": "source-branch",
            },
        ),
        (
            "tar with source tag",
            {
                "source_type": "tar",
                "source_branch": None,
                "source_tag": "test_tag",
                "source_commit": None,
                "error": "source-tag",
            },
        ),
        (
            "tar with source commit",
            {
                "source_type": "tar",
                "source_branch": None,
                "source_tag": None,
                "source_commit": "commit",
                "error": "source-commit",
            },
        ),
        (
            "deb with source branch",
            {
                "source_type": "deb",
                "source_branch": "test_branch",
                "source_tag": None,
                "source_commit": None,
                "error": "source-branch",
            },
        ),
        (
            "deb with source tag",
            {
                "source_type": "deb",
                "source_branch": None,
                "source_tag": "test_tag",
                "source_commit": None,
                "error": "source-tag",
            },
        ),
        (
            "deb with source commit",
            {
                "source_type": "deb",
                "source_branch": None,
                "source_tag": None,
                "source_commit": "commit",
                "error": "source-commit",
            },
        ),
    ]

    def test(self, source_type, source_branch, source_tag, source_commit, error):
        handler = sources.get_source_handler(
            "https://source.com", source_type=source_type
        )

        with pytest.raises(sources.errors.SnapcraftSourceInvalidOptionError) as err:
            handler(
                "https://source.com",
                source_dir=".",
                source_branch=source_branch,
                source_tag=source_tag,
                source_commit=source_commit,
            )

        assert err.value.source_type == source_type
        assert err.value.option == error


class TestSourceWithBranchAndTagErrors:

    scenarios = [
        (
            "git with source branch and tag",
            {"source_type": "git", "source_branch": "test_branch", "source_tag": "tag"},
        ),
        (
            "hg with source branch and tag",
            {
                "source_type": "mercurial",
                "source_branch": "test_branch",
                "source_tag": "tag",
            },
        ),
    ]

    def test(self, source_type, source_branch, source_tag):
        handler = sources.get_source_handler(
            "https://source.com", source_type=source_type
        )

        with pytest.raises(
            sources.errors.SnapcraftSourceIncompatibleOptionsError
        ) as error:
            handler(
                "https://source.com",
                source_dir=".",
                source_branch=source_branch,
                source_tag=source_tag,
            )

        assert error.value.source_type == source_type
        assert error.value.options == ["source-tag", "source-branch"]


def test_get(tmp_work_path):
    file_path = tmp_work_path / "file"
    file_path.touch()

    class Options:
        source = "."

    sources.get("src", "useless-arg", Options())

    src_dir = tmp_work_path / "src"
    assert src_dir.is_dir()
    assert (src_dir / "file").is_file()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.plugins.v2.qmake import QMakePlugin


def test_schema():
    assert QMakePlugin.get_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "type": "object",
        "additionalProperties": False,
        "properties": {
            "qmake-parameters": {
                "type": "array",
                "uniqueItems": True,
                "items": {"type": "string"},
                "default": [],
            },
            "qmake-project-file": {"type": "string", "default": ""},
        },
    }


def test_get_build_packages():
    class Options:
        qmake_parameters = list()
        qmake_project_file = ""

    plugin = QMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_packages() == {"g++", "make", "qt5-qmake"}


def test_get_build_snaps():
    class Options:
        qmake_parameters = list()
        qmake_project_file = ""

    plugin = QMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_snaps() == set()


def test_get_build_environment():
    class Options:
        qmake_parameters = list()
        qmake_project_file = ""

    plugin = QMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_environment() == {"QT_SELECT": "qt5"}


def test_out_of_source_build():
    class Options:
        qmake_parameters = list()
        qmake_project_file = ""

    plugin = QMakePlugin(part_name="my-part", options=Options())

    assert plugin.out_of_source_build is True


def test_get_build_commands():
    class Options:
        qmake_parameters = list()
        qmake_project_file = ""

    plugin = QMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_commands() == [
        'qmake QMAKE_CFLAGS+="${CFLAGS:-}" QMAKE_CXXFLAGS+="${CXXFLAGS:-}" QMAKE_LFLAGS+="${LDFLAGS:-}" "${SNAPCRAFT_PART_SRC_WORK}"',
        'env -u CFLAGS -u CXXFLAGS make -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
        'make install INSTALL_ROOT="${SNAPCRAFT_PART_INSTALL}"',
    ]


def test_get_build_commands_with_qmake_parameters():
    class Options:
        qmake_parameters = ["-Wall"]
        qmake_project_file = ""

    plugin = QMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_commands() == [
        'qmake QMAKE_CFLAGS+="${CFLAGS:-}" QMAKE_CXXFLAGS+="${CXXFLAGS:-}" QMAKE_LFLAGS+="${LDFLAGS:-}" -Wall "${SNAPCRAFT_PART_SRC_WORK}"',
        'env -u CFLAGS -u CXXFLAGS make -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
        'make install INSTALL_ROOT="${SNAPCRAFT_PART_INSTALL}"',
    ]


def test_get_build_commands_with_qmake_project_file():
    class Options:
        qmake_parameters = list()
        qmake_project_file = "foo.pro"

    plugin = QMakePlugin(part_name="my-part", options=Options())

    assert plugin.get_build_commands() == [
        'qmake QMAKE_CFLAGS+="${CFLAGS:-}" QMAKE_CXXFLAGS+="${CXXFLAGS:-}" QMAKE_LFLAGS+="${LDFLAGS:-}" "${SNAPCRAFT_PART_SRC_WORK}/foo.pro"',
        'env -u CFLAGS -u CXXFLAGS make -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
        'make install INSTALL_ROOT="${SNAPCRAFT_PART_INSTALL}"',
    ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.plugins.v2.autotools import AutotoolsPlugin


def test_schema():
    assert AutotoolsPlugin.get_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "type": "object",
        "additionalProperties": False,
        "properties": {
            "autotools-configure-parameters": {
                "type": "array",
                "uniqueItems": True,
                "items": {"type": "string"},
                "default": [],
            }
        },
    }


def test_get_build_packages():
    plugin = AutotoolsPlugin(part_name="my-part", options=lambda: None)

    assert plugin.get_build_packages() == {
        "autoconf",
        "automake",
        "autopoint",
        "gcc",
        "libtool",
    }


def test_get_build_environment():
    plugin = AutotoolsPlugin(part_name="my-part", options=lambda: None)

    assert plugin.get_build_environment() == dict()


def test_get_build_commands():
    class Options:
        autotools_configure_parameters = list()

    plugin = AutotoolsPlugin(part_name="my-part", options=Options())

    assert plugin.get_build_commands() == [
        "[ ! -f ./configure ] && [ -f ./autogen.sh ] && env NOCONFIGURE=1 ./autogen.sh",
        "[ ! -f ./configure ] && [ -f ./bootstrap ] && env NOCONFIGURE=1 ./bootstrap",
        "[ ! -f ./configure ] && autoreconf --install",
        "./configure",
        'make -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
        'make install DESTDIR="${SNAPCRAFT_PART_INSTALL}"',
    ]


def test_get_build_commands_with_configure_parameters():
    class Options:
        autotools_configure_parameters = ["--with-foo=true", "--prefix=/foo"]

    plugin = AutotoolsPlugin(part_name="my-part", options=Options())

    assert plugin.get_build_commands() == [
        "[ ! -f ./configure ] && [ -f ./autogen.sh ] && env NOCONFIGURE=1 ./autogen.sh",
        "[ ! -f ./configure ] && [ -f ./bootstrap ] && env NOCONFIGURE=1 ./bootstrap",
        "[ ! -f ./configure ] && autoreconf --install",
        "./configure --with-foo=true --prefix=/foo",
        'make -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
        'make install DESTDIR="${SNAPCRAFT_PART_INSTALL}"',
    ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

from testtools.matchers import Equals

from snapcraft.internal.meta import errors
from snapcraft.internal.meta.hooks import Hook
from tests import unit


class GenericHookTests(unit.TestCase):
    def test_empty(self):
        hook_dict = OrderedDict({})
        hook_name = "hook-test"

        hook = Hook(hook_name=hook_name)
        hook.validate()

        self.assertEqual(hook.to_dict(), hook_dict)
        self.assertEqual(hook.hook_name, hook_name)

    def test_empty_from_dict(self):
        hook_dict = OrderedDict({})
        hook_name = "hook-test"

        hook = Hook.from_dict(hook_dict=hook_dict, hook_name=hook_name)
        hook.validate()

        self.assertEqual(hook.to_dict(), hook_dict)
        self.assertEqual(hook.hook_name, hook_name)

    def test_passthrough(self):
        hook_dict = OrderedDict({"passthrough": {"otherkey": "othervalue"}})
        hook_name = "hook-test"

        hook = Hook.from_dict(hook_dict=hook_dict, hook_name=hook_name)
        hook.validate()

        transformed_dict = OrderedDict({"otherkey": "othervalue"})

        self.assertEqual(hook.to_dict(), transformed_dict)
        self.assertEqual(hook.hook_name, hook_name)
        self.assertEqual(hook.passthrough, hook_dict["passthrough"])

    def test_plugs(self):
        hook_dict = OrderedDict({"plugs": ["plug1", "plug2"]})
        hook_name = "hook-test"

        hook = Hook.from_dict(hook_dict=hook_dict, hook_name=hook_name)
        hook.validate()

        self.assertEqual(hook.to_dict(), hook_dict)
        self.assertEqual(hook.hook_name, hook_name)

    def test_command_chain(self):
        hook_dict = OrderedDict({"command-chain": ["cmd1", "cmd2"]})
        hook_name = "hook-test"

        hook = Hook.from_dict(hook_dict=hook_dict, hook_name=hook_name)
        hook.validate()

        self.assertEqual(hook.to_dict(), hook_dict)
        self.assertEqual(hook.hook_name, hook_name)

    def test_invalid_command_chain(self):
        hook_dict = OrderedDict({"command-chain": ["&/foo/bar"]})
        hook_name = "hook-test"

        hook = Hook.from_dict(hook_dict=hook_dict, hook_name=hook_name)

        error = self.assertRaises(errors.HookValidationError, hook.validate)
        self.assertThat(
            str(error),
            Equals(
                "failed to validate hook=hook-test: '&/foo/bar' is not a valid command-chain command."
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent
from unittest import mock

import fixtures

from snapcraft.cli import echo
from tests import unit


class EchoTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.click_confirm = self.useFixture(
            fixtures.MockPatch("click.confirm", return_value=True)
        )
        self.click_prompt = self.useFixture(
            fixtures.MockPatch("click.prompt", return_value=True)
        )

    @mock.patch("sys.stdin.isatty", return_value=False)
    def test_is_tty_disconnected(self, tty_mock):
        result = echo.is_tty_connected()

        self.assertEqual(result, False)

    @mock.patch("sys.stdin.isatty", return_value=True)
    def test_is_tty_connected(self, tty_mock):
        result = echo.is_tty_connected()

        self.assertEqual(result, True)

    @mock.patch("snapcraft.cli.echo.is_tty_connected", return_value=False)
    def test_echo_confirm_is_not_tty(self, tty_mock):
        echo.confirm("message")

        self.click_confirm.mock.assert_not_called()

    @mock.patch("snapcraft.cli.echo.is_tty_connected", return_value=True)
    def test_echo_confirm_is_tty(self, tty_mock):
        echo.confirm("message")

        self.click_confirm.mock.assert_called_once_with(
            "message",
            default=False,
            abort=False,
            prompt_suffix=": ",
            show_default=True,
            err=False,
        )

    @mock.patch("snapcraft.cli.echo.is_tty_connected", return_value=True)
    def test_echo_confirm_default(self, tty_mock):
        echo.confirm("message", default="the new default")

        self.click_confirm.mock.assert_called_once_with(
            "message",
            default="the new default",
            abort=False,
            prompt_suffix=": ",
            show_default=True,
            err=False,
        )

    @mock.patch("snapcraft.cli.echo.is_tty_connected", return_value=False)
    def test_echo_prompt_is_not_tty(self, tty_mock):
        echo.prompt("message")

        self.click_prompt.mock.assert_not_called()

    @mock.patch("snapcraft.cli.echo.is_tty_connected", return_value=True)
    def test_echo_prompt_is_tty(self, tty_mock):
        echo.prompt("message")

        self.click_prompt.mock.assert_called_once_with(
            "message",
            default=None,
            hide_input=False,
            confirmation_prompt=False,
            type=None,
            value_proc=None,
            prompt_suffix=": ",
            show_default=True,
            err=False,
        )

    @mock.patch("snapcraft.cli.echo.is_tty_connected", return_value=True)
    def test_echo_prompt_default(self, tty_mock):
        echo.prompt("message", default="the new default")

        self.click_prompt.mock.assert_called_once_with(
            "message",
            default="the new default",
            hide_input=False,
            confirmation_prompt=False,
            type=None,
            value_proc=None,
            prompt_suffix=": ",
            show_default=True,
            err=False,
        )


def test_exit_error_required_kwargs(mock_echo_error, mock_sys_exit):
    echo.exit_error(brief="You failed!", resolution="Try again!")

    mock_echo_error.assert_called_once_with(
        dedent(
            """\
    You failed!

    Recommended resolution:
    Try again!"""
        )
    )
    mock_sys_exit.assert_called_once_with(2)


def test_exit_error_all_kwargs(mock_echo_error, mock_sys_exit):
    echo.exit_error(
        brief="You failed!",
        resolution="Try again!",
        details="Try harder!",
        docs_url="https://snapcraft.io/failure",
        exit_code=5,
    )

    mock_echo_error.assert_called_once_with(
        dedent(
            """\
    You failed!

    Recommended resolution:
    Try again!

    Detailed information:
    Try harder!

    For more information, check out:
    https://snapcraft.io/failure"""
        )
    )
    mock_sys_exit.assert_called_once_with(5)


def test_exit_error_all_kwargs_except_details(mock_echo_error, mock_sys_exit):
    echo.exit_error(
        brief="You failed!",
        resolution="Try again!",
        docs_url="https://snapcraft.io/failure",
        exit_code=255,
    )

    mock_echo_error.assert_called_once_with(
        dedent(
            """\
    You failed!

    Recommended resolution:
    Try again!

    For more information, check out:
    https://snapcraft.io/failure"""
        )
    )
    mock_sys_exit.assert_called_once_with(255)


def test_exit_error_all_kwargs_except_docs_url(mock_echo_error, mock_sys_exit):
    echo.exit_error(
        brief="You failed!", resolution="Try again!", details="Try harder!", exit_code=7
    )

    mock_echo_error.assert_called_once_with(
        dedent(
            """\
    You failed!

    Recommended resolution:
    Try again!

    Detailed information:
    Try harder!"""
        )
    )
    mock_sys_exit.assert_called_once_with(7)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from snapcraft.internal.repo import ua_manager


def test_ua_manager(fake_process):
    ua_token = "test-ua-token"

    fake_process.register_subprocess(
        ["sudo", "ua", "status", "--format", "json"],
        stdout='{"attached": false, "_doc": "Content..."}',
    )
    fake_process.register_subprocess(["sudo", "ua", "attach", "test-ua-token"])
    fake_process.register_subprocess(["sudo", "ua", "detach", "--assume-yes"])

    with ua_manager.ua_manager(ua_token):
        pass

    assert list(fake_process.calls) == [
        ["sudo", "ua", "status", "--format", "json"],
        ["sudo", "ua", "attach", "test-ua-token"],
        ["sudo", "ua", "detach", "--assume-yes"],
    ]


def test_ua_manager_already_attached(fake_process):
    fake_process.register_subprocess(
        ["sudo", "ua", "status", "--format", "json"],
        stdout='{"attached": true, "_doc": "Content..."}',
    )

    with ua_manager.ua_manager("ua-token"):
        pass

    assert list(fake_process.calls) == [["sudo", "ua", "status", "--format", "json"]]


def test_ua_manager_detaches_on_error(fake_process):
    fake_process.register_subprocess(
        ["sudo", "ua", "status", "--format", "json"],
        stdout='{"attached": false, "_doc": "Content..."}',
    )
    fake_process.register_subprocess(["sudo", "ua", "attach", "test-ua-token"])
    fake_process.register_subprocess(["sudo", "ua", "detach", "--assume-yes"])

    with pytest.raises(RuntimeError):
        with ua_manager.ua_manager("test-ua-token"):
            raise RuntimeError("whoopsie")

    assert list(fake_process.calls) == [
        ["sudo", "ua", "status", "--format", "json"],
        ["sudo", "ua", "attach", "test-ua-token"],
        ["sudo", "ua", "detach", "--assume-yes"],
    ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import sys
import tempfile
from unittest import mock

import fixtures
import pytest
from testtools.matchers import Contains, EndsWith, Equals, NotEquals, StartsWith

from snapcraft import ProjectOptions
from snapcraft.internal import elf, errors
from tests import fixture_setup, unit


class TestElfBase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.fake_elf = fixture_setup.FakeElf(root_path=self.path)
        self.useFixture(self.fake_elf)
        self.content_dirs = []
        self.arch_triplet = ProjectOptions().arch_triplet


class TestLdLibraryPathParser(unit.TestCase):
    def _write_conf_file(self, contents):
        tmp = tempfile.NamedTemporaryFile(delete=False, mode="w")
        self.addCleanup(os.remove, tmp.name)

        tmp.write(contents)
        tmp.close()

        return tmp.name

    def test_extract_ld_library_paths(self):
        file_path = self._write_conf_file(
            """# This is a comment
/foo/bar
/colon:/separated,/comma\t/tab /space # This is another comment
/baz"""
        )

        self.assertThat(
            elf._extract_ld_library_paths(file_path),
            Equals(
                ["/foo/bar", "/colon", "/separated", "/comma", "/tab", "/space", "/baz"]
            ),
        )


class TestElfFileSmoketest(unit.TestCase):
    def test_bin_echo(self):
        # Try parsing a file without the pyelftools logic mocked out
        elf_file = elf.ElfFile(path=sys.executable)

        self.assertThat(elf_file.path, Equals(sys.executable))

        # The arch attribute will be a tuple of three strings
        self.assertTrue(isinstance(elf_file.arch, tuple))
        self.assertThat(len(elf_file.arch), Equals(3))
        self.assertThat(elf_file.arch[0], StartsWith("ELFCLASS"))
        self.assertThat(elf_file.arch[1], StartsWith("ELFDATA"))
        self.assertThat(elf_file.arch[2], StartsWith("EM_"))

        # We expect Python to be a dynamic linked executable with an
        # ELF interpreter.
        self.assertTrue(isinstance(elf_file.interp, str))
        self.assertThat(elf_file.interp, NotEquals(""))

        # Python is not a shared library, so has no soname or defined versions
        self.assertThat(elf_file.soname, Equals(""))
        self.assertThat(elf_file.versions, Equals(set()))

        # We expect that Python will be linked to libc
        for lib in elf_file.needed.values():
            if lib.name.startswith("libc.so"):
                break
        else:
            self.fail("Expected to find libc in needed library list")

        self.assertTrue(isinstance(lib.name, str))
        for version in lib.versions:
            self.assertTrue(
                isinstance(version, str), "expected {!r} to be a string".format(version)
            )

        # GCC adds a build ID to executables
        self.assertThat(elf_file.build_id, NotEquals(""))

        # If the Python interpreter is distro packaged, it probably
        # doesn't have debug info, but we don't know for sure.
        # Instead just check that it is a boolean.
        self.assertTrue(isinstance(elf_file.has_debug_info, bool))

        # Ensure type is detered as executable.
        self.assertThat(elf_file.elf_type, Equals("ET_EXEC"))


class TestInvalidElf(unit.TestCase):
    def test_invalid_elf_file(self):
        invalid_elf = os.path.join(self.path, "invalid-elf")
        open(invalid_elf, "wb").write(b"\x7fELF\x00")

        elf_files = elf.get_elf_files(self.path, ["invalid-elf"])
        self.assertThat(elf_files, Equals(set()))


class TestMissingLibraries(TestElfBase):
    def test_get_libraries_missing_libs(self):
        elf_file = self.fake_elf["fake_elf-with-missing-libs"]
        libs = elf_file.load_dependencies(
            root_path=self.fake_elf.root_path,
            core_base_path=self.fake_elf.core_base_path,
            arch_triplet=self.arch_triplet,
            content_dirs=self.content_dirs,
        )

        self.assertThat(
            libs,
            Equals(set([self.fake_elf.root_libraries["foo.so.1"], "missing.so.2"])),
        )


class TestGetLibraries(TestElfBase):
    def setUp(self):
        super().setUp()

        self.useFixture(fixtures.MockPatch("os.path.exists", return_value=True))

    def test_get_libraries(self):
        elf_file = self.fake_elf["fake_elf-2.23"]
        libs = elf_file.load_dependencies(
            root_path=self.fake_elf.root_path,
            core_base_path=self.fake_elf.core_base_path,
            arch_triplet=self.arch_triplet,
            content_dirs=self.content_dirs,
        )

        self.assertThat(
            libs,
            Equals(
                set([self.fake_elf.root_libraries["foo.so.1"], "/usr/lib/bar.so.2"])
            ),
        )

    def test_get_libraries_with_soname_cache(self):
        elf_file = self.fake_elf["fake_elf-2.23"]

        arch = ("ELFCLASS64", "ELFDATA2LSB", "EM_X86_64")
        soname_cache = elf.SonameCache()
        soname_cache[arch, "bar.so.2"] = "/lib/bar.so.2"

        libs = elf_file.load_dependencies(
            root_path=self.fake_elf.root_path,
            core_base_path=self.fake_elf.core_base_path,
            arch_triplet=self.arch_triplet,
            content_dirs=self.content_dirs,
            soname_cache=soname_cache,
        )

        # With no cache this would have returned '/usr/lib/bar.so.2'
        self.assertThat(
            libs,
            Equals(set([self.fake_elf.root_libraries["foo.so.1"], "/lib/bar.so.2"])),
        )

    def test_primed_libraries_are_preferred(self):
        elf_file = self.fake_elf["fake_elf-2.23"]
        libs = elf_file.load_dependencies(
            root_path=self.fake_elf.root_path,
            core_base_path=self.fake_elf.core_base_path,
            arch_triplet=self.arch_triplet,
            content_dirs=self.content_dirs,
        )

        self.assertThat(
            libs,
            Equals(
                frozenset(
                    [self.fake_elf.root_libraries["foo.so.1"], "/usr/lib/bar.so.2"]
                )
            ),
        )

    def test_non_elf_primed_sonames_matches_are_ignored(self):
        primed_foo = os.path.join(self.fake_elf.root_path, "foo.so.1")
        with open(primed_foo, "wb") as f:
            # A bz2 header
            f.write(b"\x42\x5a\x68")

        elf_file = self.fake_elf["fake_elf-2.23"]
        libs = elf_file.load_dependencies(
            root_path=self.fake_elf.root_path,
            core_base_path=self.fake_elf.core_base_path,
            arch_triplet=self.arch_triplet,
            content_dirs=self.content_dirs,
        )

        self.assertThat(libs, Equals(frozenset(["/lib/foo.so.1", "/usr/lib/bar.so.2"])))

    def test_get_libraries_excludes_slash_snap(self):
        elf_file = self.fake_elf["fake_elf-with-core-libs"]
        libs = elf_file.load_dependencies(
            root_path=self.fake_elf.root_path,
            core_base_path=self.fake_elf.core_base_path,
            arch_triplet=self.arch_triplet,
            content_dirs=self.content_dirs,
        )

        self.assertThat(
            libs,
            Equals(
                set([self.fake_elf.root_libraries["foo.so.1"], "/usr/lib/bar.so.2"])
            ),
        )

    def test_get_libraries_ldd_failure_logs_warning(self):
        self.fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(self.fake_logger)

        elf_file = self.fake_elf["fake_elf-bad-ldd"]
        libs = elf_file.load_dependencies(
            root_path=self.fake_elf.root_path,
            core_base_path=self.fake_elf.core_base_path,
            arch_triplet=self.arch_triplet,
            content_dirs=self.content_dirs,
        )

        self.assertThat(libs, Equals(set()))
        self.assertThat(
            self.fake_logger.output,
            Contains("Unable to determine library dependencies for"),
        )

    def test_existing_host_library_searched_for(self):
        elf_file = self.fake_elf["fake_elf-with-host-libraries"]

        class MooLibrary(elf.Library):
            """A Library implementation that always returns valid for moo."""

            def _is_valid_elf(self, resolved_path: str) -> bool:
                #  This path is defined in ldd for fake_elf-with-host-libraries.
                if resolved_path == "/usr/lib/moo.so.2":
                    return True
                else:
                    return super()._is_valid_elf(resolved_path)

        with mock.patch("snapcraft.internal.elf.Library", side_effect=MooLibrary):
            libs = elf_file.load_dependencies(
                root_path=self.fake_elf.root_path,
                core_base_path=self.fake_elf.core_base_path,
                arch_triplet=self.arch_triplet,
                content_dirs=self.content_dirs,
            )

        self.assertThat(libs, Equals({self.fake_elf.root_libraries["moo.so.2"]}))


class TestLibrary(TestElfBase):
    def test_is_valid_elf_ignores_corrupt_files(self):
        soname = "libssl.so.1.0.0"
        soname_path = os.path.join(self.path, soname)
        library = elf.Library(
            soname=soname,
            soname_path=soname_path,
            search_paths=[self.path],
            core_base_path="/snap/core/current",
            arch=("ELFCLASS64", "ELFDATA2LSB", "EM_X86_64"),
            soname_cache=elf.SonameCache(),
        )

        self.assertThat(library._is_valid_elf(soname_path), Equals(True))

        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.elf.ElfFile",
                side_effect=errors.CorruptedElfFileError(
                    path=soname_path, error=RuntimeError()
                ),
            )
        )

        self.assertThat(library._is_valid_elf(soname_path), Equals(False))


class TestGetElfFiles(TestElfBase):
    def test_get_elf_files(self):
        elf_files = elf.get_elf_files(self.fake_elf.root_path, {"fake_elf-2.23"})

        self.assertThat(len(elf_files), Equals(1))
        elf_file = set(elf_files).pop()
        self.assertThat(elf_file.interp, Equals("/lib64/ld-linux-x86-64.so.2"))

    def test_skip_object_files(self):
        open(os.path.join(self.fake_elf.root_path, "object_file.o"), "w").close()

        elf_files = elf.get_elf_files(self.fake_elf.root_path, {"object_file.o"})
        self.assertThat(elf_files, Equals(set()))

    def test_no_find_dependencies_statically_linked(self):
        elf_files = elf.get_elf_files(self.fake_elf.root_path, {"fake_elf-static"})
        self.assertThat(elf_files, Equals(set()))

    def test_elf_with_execstack(self):
        elf_files = elf.get_elf_files(
            self.fake_elf.root_path, {"fake_elf-with-execstack"}
        )
        elf_file = set(elf_files).pop()
        self.assertThat(elf_file.execstack_set, Equals(True))

    def test_elf_without_execstack(self):
        elf_files = elf.get_elf_files(self.fake_elf.root_path, {"fake_elf-2.23"})
        elf_file = set(elf_files).pop()
        self.assertThat(elf_file.execstack_set, Equals(False))

    def test_non_elf_files(self):
        with open(os.path.join(self.fake_elf.root_path, "non-elf"), "wb") as f:
            # A bz2 header
            f.write(b"\x42\x5a\x68")

        elf_files = elf.get_elf_files(self.fake_elf.root_path, {"non-elf"})
        self.assertThat(elf_files, Equals(set()))

    def test_symlinks(self):
        symlinked_path = os.path.join(self.fake_elf.root_path, "symlinked")
        os.symlink("/bin/dash", symlinked_path)

        elf_files = elf.get_elf_files(self.fake_elf.root_path, {"symlinked"})

        self.assertThat(elf_files, Equals(set()))

    def test_device_files(self):
        elf_files = elf.get_elf_files("/dev", {"null"})
        self.assertThat(elf_files, Equals(set()))

    def test_fifo(self):
        fifo_path = os.path.join(self.fake_elf.root_path, "fifo")
        os.mkfifo(fifo_path)

        elf_files = elf.get_elf_files(self.fake_elf.root_path, {"fifo"})
        self.assertThat(elf_files, Equals(set()))


class TestGetRequiredGLIBC(TestElfBase):
    def setUp(self):
        super().setUp()

        self.elf_file = self.fake_elf["fake_elf-2.23"]

    def test_get_required_glibc(self):
        self.assertThat(self.elf_file.get_required_glibc(), Equals("2.23"))

    def test_linker_version_greater_than_required_glibc(self):
        self.assertTrue(self.elf_file.is_linker_compatible(linker_version="2.26"))

    def test_linker_version_equals_required_glibc(self):
        self.assertTrue(self.elf_file.is_linker_compatible(linker_version="2.23"))

    def test_linker_version_less_than_required_glibc(self):
        self.assertFalse(self.elf_file.is_linker_compatible(linker_version="1.2"))


class TestElfFileAttrs(TestElfBase):
    def setUp(self):
        super().setUp()

    def test_executable(self):
        elf_file = self.fake_elf["fake_elf-2.23"]

        self.assertThat(elf_file.interp, Equals("/lib64/ld-linux-x86-64.so.2"))
        self.assertThat(elf_file.soname, Equals(""))
        self.assertThat(sorted(elf_file.needed.keys()), Equals(["libc.so.6"]))

        glibc = elf_file.needed["libc.so.6"]
        self.assertThat(glibc.name, Equals("libc.so.6"))
        self.assertThat(glibc.versions, Equals({"GLIBC_2.2.5", "GLIBC_2.23"}))

    def test_shared_object(self):
        # fake_elf-shared-object has no GLIBC dependency, but two symbols
        # nonetheless
        elf_file = self.fake_elf["fake_elf-shared-object"]

        self.assertThat(elf_file.interp, Equals(""))
        self.assertThat(elf_file.soname, Equals("libfake_elf.so.0"))
        self.assertThat(sorted(elf_file.needed.keys()), Equals(["libssl.so.1.0.0"]))

        openssl = elf_file.needed["libssl.so.1.0.0"]
        self.assertThat(openssl.name, Equals("libssl.so.1.0.0"))
        self.assertThat(openssl.versions, Equals({"OPENSSL_1.0.0"}))

    def test_libssl(self):
        # libssl.so.1.0.0 defines some symbol versions it provides
        elf_file = self.fake_elf["libssl.so.1.0.0"]

        self.assertThat(elf_file.interp, Equals(""))
        self.assertThat(elf_file.soname, Equals("libssl.so.1.0.0"))
        self.assertThat(elf_file.versions, Equals({"libssl.so.1.0.0", "OPENSSL_1.0.0"}))


class TestPatcher(TestElfBase):
    def test_patch(self):
        elf_file = self.fake_elf["fake_elf-2.23"]
        # The base_path does not matter here as there are not files to
        # be crawled for.
        elf_patcher = elf.Patcher(dynamic_linker="/lib/fake-ld", root_path="/fake")
        elf_patcher.patch(elf_file=elf_file)

    def test_patch_does_nothing_if_no_interpreter(self):
        elf_file = self.fake_elf["fake_elf-static"]
        # The base_path does not matter here as there are not files to
        # be crawled for.
        elf_patcher = elf.Patcher(dynamic_linker="/lib/fake-ld", root_path="/fake")
        elf_patcher.patch(elf_file=elf_file)


class TestPatcherErrors(TestElfBase):
    def test_patch_fails_raises_patcherror_exception(self):
        elf_file = self.fake_elf["fake_elf-bad-patchelf"]
        # The base_path does not matter here as there are not files to
        # be crawled for.
        elf_patcher = elf.Patcher(dynamic_linker="/lib/fake-ld", root_path="/fake")

        self.assertRaises(
            errors.PatcherGenericError, elf_patcher.patch, elf_file=elf_file
        )


class TestSonameCache(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.arch = ("ELFCLASS64", "ELFDATA2LSB", "EM_X86_64")
        self.soname_cache = elf.SonameCache()

    def test_add_and_retrieve_soname_path(self):
        self.soname_cache[self.arch, "soname.so"] = "/fake/path/soname.so"
        self.assertThat(
            self.soname_cache[self.arch, "soname.so"], Equals("/fake/path/soname.so")
        )

    def test_add_and_verify_soname_path(self):
        self.soname_cache[self.arch, "soname.so"] = "/fake/path/soname.so"
        self.assertTrue((self.arch, "soname.so") in self.soname_cache)

    def test_reset_except_root(self):
        self.soname_cache[self.arch, "soname.so"] = "/fake/path/soname.so"
        self.soname_cache[self.arch, "soname2.so"] = "/keep/me/soname2.so"
        self.soname_cache[self.arch, "notfound.so"] = None

        self.assertTrue((self.arch, "soname.so") in self.soname_cache)
        self.assertTrue((self.arch, "soname2.so") in self.soname_cache)
        self.assertTrue((self.arch, "notfound.so") in self.soname_cache)

        self.soname_cache.reset_except_root("/keep/me")

        self.assertFalse((self.arch, "soname.so") in self.soname_cache)
        self.assertFalse((self.arch, "notfound.so") in self.soname_cache)
        self.assertTrue((self.arch, "soname2.so") in self.soname_cache)


class TestSonameCacheErrors:

    scenarios = (
        ("invalid string key", dict(key="soname.so", partial_message="The key for")),
        (
            "invalid first argument tuple",
            dict(
                key=(("ELFCLASS64", "ELFDATA2LSB"), "soname.so"),
                partial_message="The first element",
            ),
        ),
        (
            "invalid second argument type",
            dict(
                key=(("ELFCLASS64", "ELFDATA2LSB", "EM_X86_64"), 1),
                partial_message="The second element",
            ),
        ),
    )

    def test_error(self, key, partial_message):
        soname_cache = elf.SonameCache()
        with pytest.raises(EnvironmentError) as error:
            soname_cache.__setitem__(key, "/soname.so")
            assert str(error).startswith(partial_message)


# This is just a subset
_LIBC6_LIBRARIES = [
    "ld-2.26.so",
    "ld-linux-x86-64.so.2",
    "libBrokenLocale-2.26.so",
    "libBrokenLocale.so.1",
    "libSegFault.so",
    "libanl-2.26.so",
]


class HandleGlibcTestCase(unit.TestCase):
    def _setup_libc6(self):
        lib_path = os.path.join(self.path, "lib")
        libraries = {os.path.join(lib_path, l) for l in _LIBC6_LIBRARIES}

        os.mkdir(lib_path)
        for library in libraries:
            open(library, "w").close()

        return libraries

    def setUp(self):
        super().setUp()

        patcher = mock.patch("snapcraft.internal.repo.Repo.get_package_libraries")
        self.get_packages_mock = patcher.start()
        self.get_packages_mock.return_value = self._setup_libc6()
        self.addCleanup(patcher.stop)

    def test_glibc_mangling(self):
        dynamic_linker = elf.find_linker(
            root_path=self.path, snap_base_path="/snap/snap-name/current"
        )

        self.get_packages_mock.assert_called_once_with("libc6")

        self.assertThat(dynamic_linker, EndsWith("ld-2.26.so"))

    def test_bad_dynamic_linker_in_libc6_package(self):
        self.get_packages_mock.return_value = {"/usr/lib/dyn-linker-2.25.so"}
        self.assertRaises(
            RuntimeError,
            elf.find_linker,
            root_path=self.path,
            snap_base_path="/snap/snap-name/current",
        )


class TestLddParsing:
    scenarios = [
        (
            "ubuntu 20.04 basic",
            dict(
                ldd_output="""
\tlinux-vdso.so.1 (0x00007ffcae3e6000)
\tlibc.so.6 => /lib/x86_64-linux-gnu/libc.so.6 (0x00007f33eebeb000)
\t/lib64/ld-linux-x86-64.so.2 (0x00007f33eedf2000)
""",
                expected={"libc.so.6": "/lib/x86_64-linux-gnu/libc.so.6"},
            ),
        ),
        (
            "ubuntu 18.04 lspci w/o libpci",
            dict(
                ldd_output="""
\tlinux-vdso.so.1 (0x00007fffeddd1000)
\tlibpci.so.3 => not found
\tlibkmod.so.2 => /lib/x86_64-linux-gnu/libkmod.so.2 (0x00007fe500619000)
\tlibc.so.6 => /lib/x86_64-linux-gnu/libc.so.6 (0x00007fe500228000)
\t/lib64/ld-linux-x86-64.so.2 (0x00007fe500a44000)
""",
                expected={
                    "libc.so.6": "/lib/x86_64-linux-gnu/libc.so.6",
                    "libkmod.so.2": "/lib/x86_64-linux-gnu/libkmod.so.2",
                    "libpci.so.3": "libpci.so.3",
                },
            ),
        ),
        (
            "ubuntu 16.04 basic",
            dict(
                ldd_output="""
\tlinux-vdso.so.1 =>  (0x00007ffd71d64000)
\tlibc.so.6 => /lib/x86_64-linux-gnu/libc.so.6 (0x00007fda33a16000)
\t/lib64/ld-linux-x86-64.so.2 (0x00007fda33de0000)
    """,
                expected={"libc.so.6": "/lib/x86_64-linux-gnu/libc.so.6"},
            ),
        ),
        (
            "ubuntu 16.04 lspci w/o libpci",
            dict(
                ldd_output="""
\tlinux-vdso.so.1 =>  (0x00007fff305b3000)
\tlibpci.so.3 => not found
\tlibkmod.so.2 => /lib/x86_64-linux-gnu/libkmod.so.2 (0x00007faef225c000)
\tlibc.so.6 => /lib/x86_64-linux-gnu/libc.so.6 (0x00007faef1e92000)
\t/lib64/ld-linux-x86-64.so.2 (0x00007faef2473000)
    """,
                expected={
                    "libc.so.6": "/lib/x86_64-linux-gnu/libc.so.6",
                    "libkmod.so.2": "/lib/x86_64-linux-gnu/libkmod.so.2",
                    "libpci.so.3": "libpci.so.3",
                },
            ),
        ),
    ]

    def test_ldd(self, ldd_output, expected, monkeypatch, fake_process):
        monkeypatch.setattr(os.path, "exists", lambda f: True)
        monkeypatch.setattr(os.path, "abspath", lambda p: p)

        fake_process.register_subprocess(
            ["/usr/bin/ldd", "/bin/foo"], stdout=ldd_output.encode()
        )

        libraries = elf._determine_libraries(path="/bin/foo", ld_library_paths=[])

        assert libraries == expected

    def test_ldd_with_preload(
        self, ldd_output, expected, monkeypatch, fake_process, tmp_path
    ):
        libc_path = tmp_path / "libc.so.6"
        libc_path.write_text("")

        monkeypatch.setattr(os.path, "exists", lambda f: True)
        monkeypatch.setattr(os.path, "abspath", lambda p: p)
        monkeypatch.setattr(elf, "_get_host_libc_path", lambda: libc_path)

        fake_process.register_subprocess(["/usr/bin/ldd", "/bin/foo"], returncode=1)
        fake_process.register_subprocess(
            ["/usr/bin/ldd", "/bin/foo"], stdout=ldd_output.encode()
        )

        libraries = elf._determine_libraries(path="/bin/foo", ld_library_paths=[])

        assert libraries == expected

    def test_ld_trace(self, ldd_output, expected, monkeypatch, fake_process):
        monkeypatch.setattr(os.path, "exists", lambda f: True)
        monkeypatch.setattr(os.path, "abspath", lambda p: p)

        fake_process.register_subprocess(["/usr/bin/ldd", "/bin/foo"], returncode=1)
        fake_process.register_subprocess(["/usr/bin/ldd", "/bin/foo"], returncode=1)
        fake_process.register_subprocess(["/bin/foo"], stdout=ldd_output.encode())

        libraries = elf._determine_libraries(path="/bin/foo", ld_library_paths=[])

        assert libraries == expected


def test_ldd_with_preload_no_libc(monkeypatch, fake_process, tmp_path):
    monkeypatch.setattr(os.path, "exists", lambda f: True)
    monkeypatch.setattr(os.path, "abspath", lambda p: p)
    monkeypatch.setattr(elf, "_get_host_libc_path", lambda: tmp_path / "does-not-exist")

    fake_process.register_subprocess(["/usr/bin/ldd", "/bin/foo"], returncode=1)
    fake_process.register_subprocess(
        ["/bin/foo"], stdout="\tlibx.so.0 => /lib/libx.so (0x0123)".encode()
    )

    libraries = elf._determine_libraries(path="/bin/foo", ld_library_paths=[])

    assert libraries == {"libx.so.0": "/lib/libx.so"}


def test_ld_trace_os_error_for_wrong_arch(monkeypatch, fake_process, tmp_path):
    monkeypatch.setattr(os.path, "exists", lambda f: True)
    monkeypatch.setattr(os.path, "abspath", lambda p: p)
    monkeypatch.setattr(elf, "_get_host_libc_path", lambda: tmp_path / "does-not-exist")

    fake_process.register_subprocess(["/usr/bin/ldd", "/bin/foo"], returncode=1)

    def raise_os_error(*args, **kwargs):
        raise OSError(8, "Exec format error:", "does-not-exist")

    fake_process.register_subprocess(["/bin/foo"], callback=raise_os_error)

    libraries = elf._determine_libraries(path="/bin/foo", ld_library_paths=[])

    assert libraries == {}



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import fileinput
import functools
import logging
import os
import pathlib
import re
import subprocess
import sys
import tempfile
from typing import Dict, List, Optional, Set, Tuple  # noqa: F401

from xdg import BaseDirectory

from snapcraft import file_utils
from snapcraft.internal.indicators import is_dumb_terminal

from . import errors
from ._base import BaseRepo, get_pkg_name_parts
from .deb_package import DebPackage

if sys.platform == "linux":
    # Ensure importing works on non-Linux.
    from .apt_cache import AptCache

logger = logging.getLogger(__name__)

_DEB_CACHE_DIR: pathlib.Path = pathlib.Path(
    BaseDirectory.save_cache_path("snapcraft", "download")
)
_STAGE_CACHE_DIR: pathlib.Path = pathlib.Path(
    BaseDirectory.save_cache_path("snapcraft", "stage-packages")
)

_HASHSUM_MISMATCH_PATTERN = re.compile(r"(E:Failed to fetch.+Hash Sum mismatch)+")
_DEFAULT_FILTERED_STAGE_PACKAGES: List[str] = [
    "adduser",
    "apt",
    "apt-utils",
    "base-files",
    "base-passwd",
    "bash",
    "bsdutils",
    "coreutils",
    "dash",
    "debconf",
    "debconf-i18n",
    "debianutils",
    "diffutils",
    "dmsetup",
    "dpkg",
    "e2fslibs",
    "e2fsprogs",
    "file",
    "findutils",
    "gcc-4.9-base",
    "gcc-5-base",
    "gnupg",
    "gpgv",
    "grep",
    "gzip",
    "hostname",
    "init",
    "initscripts",
    "insserv",
    "libacl1",
    "libapparmor1",
    "libapt",
    "libapt-inst1.5",
    "libapt-pkg4.12",
    "libattr1",
    "libaudit-common",
    "libaudit1",
    "libblkid1",
    "libbz2-1.0",
    "libc-bin",
    "libc6",
    "libcap2",
    "libcap2-bin",
    "libcomerr2",
    "libcryptsetup4",
    "libdb5.3",
    "libdebconfclient0",
    "libdevmapper1.02.1",
    "libgcc1",
    "libgcrypt20",
    "libgpg-error0",
    "libgpm2",
    "libkmod2",
    "liblocale-gettext-perl",
    "liblzma5",
    "libmagic1",
    "libmount1",
    "libncurses5",
    "libncursesw5",
    "libpam-modules",
    "libpam-modules-bin",
    "libpam-runtime",
    "libpam0g",
    "libpcre3",
    "libprocps3",
    "libreadline6",
    "libselinux1",
    "libsemanage-common",
    "libsemanage1",
    "libsepol1",
    "libslang2",
    "libsmartcols1",
    "libss2",
    "libstdc++6",
    "libsystemd0",
    "libtext-charwidth-perl",
    "libtext-iconv-perl",
    "libtext-wrapi18n-perl",
    "libtinfo5",
    "libudev1",
    "libusb-0.1-4",
    "libustr-1.0-1",
    "libuuid1",
    "locales",
    "login",
    "lsb-base",
    "makedev",
    "manpages",
    "manpages-dev",
    "mawk",
    "mount",
    "multiarch-support",
    "ncurses-base",
    "ncurses-bin",
    "passwd",
    "perl-base",
    "procps",
    "readline-common",
    "sed",
    "sensible-utils",
    "systemd",
    "systemd-sysv",
    "sysv-rc",
    "sysvinit-utils",
    "tar",
    "tzdata",
    "ubuntu-keyring",
    "udev",
    "util-linux",
    "zlib1g",
]


IGNORE_FILTERS: Dict[str, Set[str]] = {
    "core20": {
        "python3-attr",
        "python3-blinker",
        "python3-certifi",
        "python3-cffi-backend",
        "python3-chardet",
        "python3-configobj",
        "python3-cryptography",
        # Rely on setuptools installed by plugin or found in base, unless
        # explicitly requested.
        # "python3-distutils"
        "python3-idna",
        "python3-importlib-metadata",
        "python3-jinja2",
        "python3-json-pointer",
        "python3-jsonpatch",
        "python3-jsonschema",
        "python3-jwt",
        "python3-lib2to3",
        "python3-markupsafe",
        # Provides /usr/bin/python3, don't bring in unless explicitly requested.
        # "python3-minimal"
        "python3-more-itertools",
        "python3-netifaces",
        "python3-oauthlib",
        # Rely on version brought in by setuptools, unless explicitly requested.
        # "python3-pkg-resources"
        "python3-pyrsistent",
        "python3-pyudev",
        "python3-requests",
        "python3-requests-unixsocket",
        "python3-serial",
        # Rely on version installed by plugin or found in base, unless
        # explicitly requested.
        # "python3-setuptools"
        "python3-six",
        "python3-urllib3",
        "python3-urwid",
        "python3-yaml",
        "python3-zipp",
    }
}


@functools.lru_cache(maxsize=256)
def _run_dpkg_query_search(file_path: str) -> str:
    try:
        output = (
            subprocess.check_output(
                ["dpkg-query", "-S", os.path.join(os.path.sep, file_path)],
                stderr=subprocess.STDOUT,
                env=dict(LANG="C.UTF-8"),
            )
            .decode()
            .strip()
        )
    except subprocess.CalledProcessError as call_error:
        logger.debug(
            "Error finding package for {}: {}".format(file_path, str(call_error))
        )
        raise errors.FileProviderNotFound(file_path=file_path) from call_error

    # Remove diversions
    provides_output = [p for p in output.splitlines() if not p.startswith("diversion")][
        0
    ]
    return provides_output.split(":")[0]


@functools.lru_cache(maxsize=256)
def _run_dpkg_query_list_files(package_name: str) -> Set[str]:
    output = (
        subprocess.check_output(["dpkg", "-L", package_name])
        .decode(sys.getfilesystemencoding())
        .strip()
        .split()
    )

    return {i for i in output if ("lib" in i and os.path.isfile(i))}


def _get_dpkg_list_path(base: str) -> pathlib.Path:
    return pathlib.Path(f"/snap/{base}/current/usr/share/snappy/dpkg.list")


def _get_filtered_stage_package_names(
    *, base: str, package_list: List[DebPackage]
) -> Set[str]:
    """Get filtered packages by name only - no version or architectures."""
    manifest_packages = [p.name for p in get_packages_in_base(base=base)]
    stage_packages = [p.name for p in package_list]

    return (
        set(manifest_packages) - set(stage_packages) - IGNORE_FILTERS.get(base, set())
    )


def get_packages_in_base(*, base: str) -> List[DebPackage]:
    # We do not want to break what we already have.
    if base in ("core", "core16", "core18"):
        return [DebPackage.from_unparsed(p) for p in _DEFAULT_FILTERED_STAGE_PACKAGES]

    base_package_list_path = _get_dpkg_list_path(base)
    if not base_package_list_path.exists():
        return list()

    # Lines we care about in dpkg.list had the following format:
    # ii adduser 3.118ubuntu1 all add and rem
    package_list = list()
    with fileinput.input(str(base_package_list_path)) as fp:
        for line in fp:
            if not line.startswith("ii "):
                continue
            package = DebPackage.from_unparsed(line.split()[1])
            package_list.append(package)

    return package_list


class Ubuntu(BaseRepo):
    @classmethod
    def get_package_libraries(cls, package_name: str) -> Set[str]:
        return _run_dpkg_query_list_files(package_name)

    @classmethod
    def get_package_for_file(cls, file_path: str) -> str:
        return _run_dpkg_query_search(file_path)

    @classmethod
    def get_packages_for_source_type(cls, source_type):
        if source_type == "bzr":
            packages = {"bzr"}
        elif source_type == "git":
            packages = {"git"}
        elif source_type == "tar":
            packages = {"tar"}
        elif source_type == "hg" or source_type == "mercurial":
            packages = {"mercurial"}
        elif source_type == "subversion" or source_type == "svn":
            packages = {"subversion"}
        elif source_type == "rpm2cpio":
            packages = {"rpm2cpio"}
        elif source_type == "7zip":
            packages = {"p7zip-full"}
        else:
            packages = set()

        return packages

    @classmethod
    def refresh_build_packages(cls) -> None:
        try:
            cmd = ["sudo", "--preserve-env", "apt-get", "update"]
            logger.debug(f"Executing: {cmd!r}")
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError as call_error:
            raise errors.CacheUpdateFailedError(
                "failed to run apt update"
            ) from call_error

    @classmethod
    def _check_if_all_packages_installed(cls, package_names: List[str]) -> bool:
        """Check if all given packages are installed.

        Will check versions if using <pkg_name>=<pkg_version> syntax parsed by
        get_pkg_name_parts().  Used as an optimization to skip installation
        and cache refresh if dependencies are already satisfied.

        :return True if _all_ packages are installed (with correct versions).
        """

        with AptCache() as apt_cache:
            for package in package_names:
                pkg_name, pkg_version = get_pkg_name_parts(package)
                installed_version = apt_cache.get_installed_version(
                    pkg_name, resolve_virtual_packages=True
                )

                if installed_version is None or (
                    pkg_version is not None and installed_version != pkg_version
                ):
                    return False

        return True

    @classmethod
    def _get_packages_marked_for_installation(
        cls, package_names: List[str]
    ) -> List[Tuple[str, str]]:
        with AptCache() as apt_cache:
            try:
                apt_cache.mark_packages(set(package_names))
            except errors.PackageNotFoundError as error:
                raise errors.BuildPackageNotFoundError(error.package_name)

            return apt_cache.get_packages_marked_for_installation()

    @classmethod
    def install_build_packages(cls, package_names: List[str]) -> List[str]:
        """Install packages on the host required to build.

        :param package_names: a list of package names to install.
        :type package_names: a list of strings.
        :return: a list with the packages installed and their versions.
        :rtype: list of strings.
        :raises snapcraft.repo.errors.BuildPackageNotFoundError:
            if one of the packages was not found.
        :raises snapcraft.repo.errors.PackageBrokenError:
            if dependencies for one of the packages cannot be resolved.
        :raises snapcraft.repo.errors.BuildPackagesNotInstalledError:
            if installing the packages on the host failed.
        """
        install_required = False
        package_names = sorted(package_names)

        logger.debug(f"Requested build-packages: {package_names!r}")

        # Ensure we have an up-to-date cache first if we will have to
        # install anything.
        if not cls._check_if_all_packages_installed(package_names):
            install_required = True
            cls.refresh_build_packages()

        marked_packages = cls._get_packages_marked_for_installation(package_names)
        packages = [f"{name}={version}" for name, version in sorted(marked_packages)]

        if install_required:
            cls._install_packages(packages)
        else:
            logger.debug(f"Requested build-packages already installed: {packages!r}")

        return packages

    @classmethod
    def _install_packages(cls, package_names: List[str]) -> None:
        logger.info("Installing build dependencies: %s", " ".join(package_names))
        env = os.environ.copy()
        env.update(
            {
                "DEBIAN_FRONTEND": "noninteractive",
                "DEBCONF_NONINTERACTIVE_SEEN": "true",
                "DEBIAN_PRIORITY": "critical",
            }
        )

        apt_command = [
            "sudo",
            "--preserve-env",
            "apt-get",
            "--no-install-recommends",
            "-y",
            "--allow-downgrades",
        ]
        if not is_dumb_terminal():
            apt_command.extend(["-o", "Dpkg::Progress-Fancy=1"])
        apt_command.append("install")

        try:
            subprocess.check_call(apt_command + package_names, env=env)
        except subprocess.CalledProcessError:
            raise errors.BuildPackagesNotInstalledError(packages=package_names)

        versionless_names = [get_pkg_name_parts(p)[0] for p in package_names]
        try:
            subprocess.check_call(
                ["sudo", "apt-mark", "auto"] + versionless_names, env=env
            )
        except subprocess.CalledProcessError as e:
            logger.warning(
                "Impossible to mark packages as auto-installed: {}".format(e)
            )

    @classmethod
    def fetch_stage_packages(
        cls,
        *,
        package_names: List[str],
        base: str,
        stage_packages_path: pathlib.Path,
        target_arch: str,
    ) -> List[str]:
        logger.debug(f"Requested stage-packages: {sorted(package_names)!r}")

        installed: Set[str] = set()

        package_list = [DebPackage.from_unparsed(name) for name in package_names]
        filtered_names = _get_filtered_stage_package_names(
            base=base, package_list=package_list
        )

        stage_packages_path.mkdir(exist_ok=True)
        with AptCache(
            stage_cache=_STAGE_CACHE_DIR, stage_cache_arch=target_arch
        ) as apt_cache:
            apt_cache.update()
            apt_cache.mark_packages(set(package_names))
            apt_cache.unmark_packages(filtered_names)
            for pkg_name, pkg_version, dl_path in apt_cache.fetch_archives(
                _DEB_CACHE_DIR
            ):
                logger.debug(f"Extracting stage package: {pkg_name}")
                installed.add(f"{pkg_name}={pkg_version}")
                file_utils.link_or_copy(
                    str(dl_path), str(stage_packages_path / dl_path.name)
                )

        return sorted(installed)

    @classmethod
    def unpack_stage_packages(
        cls, *, stage_packages_path: pathlib.Path, install_path: pathlib.Path
    ) -> None:
        stage_packages = stage_packages_path.glob("*.deb")

        if not stage_packages:
            return

        for pkg_path in stage_packages:
            with tempfile.TemporaryDirectory(suffix="deb-extract") as extract_dir:
                # Extract deb package.
                cls._extract_deb(pkg_path, extract_dir)
                # Mark source of files.
                marked_name = cls._extract_deb_name_version(pkg_path)
                cls._mark_origin_stage_package(extract_dir, marked_name)
                # Stage files to install_dir.
                file_utils.link_or_copy_tree(extract_dir, install_path.as_posix())
        cls.normalize(str(install_path))

    @classmethod
    def build_package_is_valid(cls, package_name) -> bool:
        with AptCache() as apt_cache:
            return apt_cache.is_package_valid(package_name)

    @classmethod
    def is_package_installed(cls, package_name) -> bool:
        with AptCache() as apt_cache:
            return apt_cache.get_installed_version(package_name) is not None

    @classmethod
    def get_installed_packages(cls) -> List[str]:
        with AptCache() as apt_cache:
            return [
                f"{pkg_name}={pkg_version}"
                for pkg_name, pkg_version in apt_cache.get_installed_packages().items()
            ]

    @classmethod
    def _extract_deb_name_version(cls, deb_path: pathlib.Path) -> str:
        try:
            output = subprocess.check_output(
                ["dpkg-deb", "--show", "--showformat=${Package}=${Version}", deb_path]
            )
        except subprocess.CalledProcessError:
            raise errors.UnpackError(deb_path)

        return output.decode().strip()

    @classmethod
    def _extract_deb(cls, deb_path: pathlib.Path, extract_dir: str) -> None:
        """Extract deb and return `<package-name>=<version>`."""
        try:
            subprocess.check_call(["dpkg-deb", "--extract", deb_path, extract_dir])
        except subprocess.CalledProcessError:
            raise errors.UnpackError(deb_path)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
from urllib.request import urlretrieve

from progressbar import AnimatedMarker, Bar, Percentage, ProgressBar, UnknownLength


def _init_progress_bar(total_length, destination, message=None):
    if not message:
        message = "Downloading {!r}".format(os.path.basename(destination))

    valid_length = total_length and total_length > 0

    if valid_length and is_dumb_terminal():
        widgets = [message, " ", Percentage()]
        maxval = total_length
    elif valid_length and not is_dumb_terminal():
        widgets = [message, Bar(marker="=", left="[", right="]"), " ", Percentage()]
        maxval = total_length
    elif not valid_length and is_dumb_terminal():
        widgets = [message]
        maxval = UnknownLength
    else:
        widgets = [message, AnimatedMarker()]
        maxval = UnknownLength

    return ProgressBar(widgets=widgets, maxval=maxval)


def download_requests_stream(request_stream, destination, message=None, total_read=0):
    """This is a facility to download a request with nice progress bars."""

    # Doing len(request_stream.content) may defeat the purpose of a
    # progress bar
    total_length = 0
    if not request_stream.headers.get("Content-Encoding", ""):
        total_length = int(request_stream.headers.get("Content-Length", "0"))
        # Content-Length in the case of resuming will be
        # Content-Length - total_read so we add back up to have the feel of
        # resuming
        if os.path.exists(destination):
            total_length += total_read

    progress_bar = _init_progress_bar(total_length, destination, message)
    progress_bar.start()

    if os.path.exists(destination):
        mode = "ab"
    else:
        mode = "wb"
    with open(destination, mode) as destination_file:
        for buf in request_stream.iter_content(1024):
            destination_file.write(buf)
            if not is_dumb_terminal():
                total_read += len(buf)
                progress_bar.update(total_read)
    progress_bar.finish()


class UrllibDownloader(object):
    """This is a facility to download an uri with nice progress bars."""

    def __init__(self, uri, destination, message=None):
        self.uri = uri
        self.destination = destination
        self.message = message
        self.progress_bar = None

    def download(self):
        urlretrieve(self.uri, self.destination, self._progress_callback)

        if self.progress_bar:
            self.progress_bar.finish()

    def _progress_callback(self, block_num, block_size, total_length):
        if not self.progress_bar:
            self.progress_bar = _init_progress_bar(
                total_length, self.destination, self.message
            )
            self.progress_bar.start()

        total_read = block_num * block_size
        self.progress_bar.update(
            min(total_read, total_length) if total_length > 0 else total_read
        )


def download_urllib_source(uri, destination, message=None):
    UrllibDownloader(uri, destination, message).download()


def is_dumb_terminal():
    """Return True if on a dumb terminal."""
    is_stdout_tty = os.isatty(1)
    is_term_dumb = os.environ.get("TERM", "") == "dumb"
    return not is_stdout_tty or is_term_dumb



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import os
import pathlib
import subprocess
import sys
from typing import Dict, List, Set

import click
from catkin_pkg import packages as catkin_packages

from snapcraft.internal.repo import Repo
from snapcraft.plugins.v1._ros.rosdep import _parse_rosdep_resolve_dependencies
from snapcraft.plugins.v2 import PluginV2


class RosPlugin(PluginV2):
    """Base class for ROS-related plugins. Not intended for use by end users."""

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return {
            "python3-rosdep",
        }

    def get_build_environment(self) -> Dict[str, str]:
        return {
            "ROS_PYTHON_VERSION": "3",
        }

    @property
    def out_of_source_build(self):
        return True

    def _get_workspace_activation_commands(self) -> List[str]:
        """Return a list of commands source a ROS workspace.

        The commands returned will be run before doing anything else.
        They will be run in a single shell instance with the rest of
        the build step, so these commands can affect the commands that
        follow.

        snapcraftctl can be used in the script to call out to snapcraft
        specific functionality.
        """

        # There are a number of unbound vars, disable flag
        # after saving current state to restore after.
        return [
            'state="$(set +o)"',
            "set +u",
            ". /opt/ros/$ROS_DISTRO/setup.sh",
            'eval "$(state)"',
        ]

    @abc.abstractmethod
    def _get_build_commands(self) -> List[str]:
        """Return a list of commands to run during the build step.

        The commands returned will be run after rosdep is used to install
        build dependencies, and before staging runtime dependencies.
        These commands are run in a single shell instance with the rest
        of the build step, so these commands can be affected by the commands
        preceding it, and can affect those that follow.

        snapcraftctl can be used in the script to call out to snapcraft
        specific functionality.
        """

    def _get_stage_runtime_dependencies_commands(self) -> List[str]:
        env = dict(LANG="C.UTF-8", LC_ALL="C.UTF-8")

        for key in [
            "PATH",
            "SNAP",
            "SNAP_ARCH",
            "SNAP_NAME",
            "SNAP_VERSION",
            "http_proxy",
            "https_proxy",
        ]:
            if key in os.environ:
                env[key] = os.environ[key]

        env_flags = [f"{key}={value}" for key, value in env.items()]
        return [
            " ".join(
                [
                    "env",
                    "-i",
                    *env_flags,
                    sys.executable,
                    "-I",
                    os.path.abspath(__file__),
                    "stage-runtime-dependencies",
                    "--part-src",
                    "$SNAPCRAFT_PART_SRC",
                    "--part-install",
                    "$SNAPCRAFT_PART_INSTALL",
                    "--ros-distro",
                    "$ROS_DISTRO",
                    "--target-arch",
                    "$SNAPCRAFT_TARGET_ARCH",
                ]
            )
        ]

    def get_build_commands(self) -> List[str]:
        return (
            self._get_workspace_activation_commands()
            + [
                "if [ ! -f /etc/ros/rosdep/sources.list.d/20-default.list ]; then sudo rosdep init; fi",
                "rosdep update --include-eol-distros --rosdistro $ROS_DISTRO",
                "rosdep install --default-yes --ignore-packages-from-source --from-paths $SNAPCRAFT_PART_SRC",
            ]
            + self._get_build_commands()
            + self._get_stage_runtime_dependencies_commands()
        )


@click.group()
def plugin_cli():
    pass


@plugin_cli.command()
@click.option("--part-src", envvar="SNAPCRAFT_PART_SRC", required=True)
@click.option("--part-install", envvar="SNAPCRAFT_PART_INSTALL", required=True)
@click.option("--ros-distro", envvar="ROS_DISTRO", required=True)
@click.option("--target-arch", envvar="SNAPCRAFT_TARGET_ARCH", required=True)
def stage_runtime_dependencies(
    part_src: str, part_install: str, ros_distro: str, target_arch: str
):
    click.echo("Staging runtime dependencies...")
    # TODO: support python packages (only apt currently supported)
    apt_packages: Set[str] = set()

    for pkg in catkin_packages.find_packages(part_src).values():
        for dep in pkg.exec_depends:
            cmd = ["rosdep", "resolve", dep.name, "--rosdistro", ros_distro]
            try:
                click.echo(f"Running {cmd!r}")
                proc = subprocess.run(
                    cmd,
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    env=dict(PATH=os.environ["PATH"]),
                )
            except subprocess.CalledProcessError as error:
                click.echo(f"failed to run {cmd!r}: {error.output}")

            parsed = _parse_rosdep_resolve_dependencies(
                dep, proc.stdout.decode().strip()
            )
            apt_packages |= parsed.pop("apt", set())

            if parsed:
                click.echo(f"unhandled dependencies: {parsed!r}")

    if apt_packages:
        package_names = sorted(apt_packages)
        install_path = pathlib.Path(part_install)
        stage_packages_path = install_path.parent / "stage_packages"

        click.echo(f"Fetching stage packages: {package_names!r}")
        Repo.fetch_stage_packages(
            package_names=package_names,
            base="core20",
            stage_packages_path=stage_packages_path,
            target_arch=target_arch,
        )

        click.echo(f"Unpacking stage packages: {package_names!r}")
        Repo.unpack_stage_packages(
            stage_packages_path=stage_packages_path, install_path=install_path
        )


if __name__ == "__main__":
    plugin_cli()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
from typing import Any, Callable
from unittest import skipUnless

from tests import os_release


def skip_unless_codename(codename, message: str) -> Callable[..., Callable[..., None]]:
    if type(codename) is str:
        codename = [codename]

    def _wrap(func: Callable[..., None]) -> Callable[..., None]:
        actual_codename = os_release.get_version_codename()

        @functools.wraps(func)
        @skipUnless(actual_codename in codename, message)
        def _skip_test(*args: Any, **kwargs: Any) -> None:
            func(*args, **kwargs)

        return _skip_test

    return _wrap



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017,2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pkgutil
from typing import Iterable

import click

import snapcraft
from snapcraft.internal import errors
from snapcraft.internal.common import format_output_in_columns, get_terminal_width
from snapcraft.project import errors as project_errors

from ._options import get_project


@click.group()
def discoverycli():
    pass


def _try_get_base_from_project() -> str:
    """Return the base used in the Snapcraft project or the current default"""
    try:
        base = get_project()._snap_meta.get_build_base()
    except (errors.ProjectNotFoundError, project_errors.MissingSnapcraftYamlError):
        base = "core20"

    return base


def _get_modules_iter(base: str) -> Iterable:
    if base in ("core", "core16", "core18"):
        modules_path = snapcraft.plugins.v1.__path__  # type: ignore  # mypy issue #1422
    else:
        modules_path = snapcraft.plugins.v2.__path__  # type: ignore  # mypy issue #1422

    # TODO make this part of plugin_finder.
    return pkgutil.iter_modules(modules_path)


@discoverycli.command("list-plugins")
@click.option(
    "--base",
    help="Show plugins for specific base",
    type=click.Choice(["core", "core16", "core18", "core20"]),
)
def list_plugins(base):
    """List the available plugins that handle different types of part.

    This command has an alias of `plugins`.
    """
    if base is None:
        base = _try_get_base_from_project()

    plugins = []
    for _, modname, _ in _get_modules_iter(base):
        # Only add non-private modules/packages to the plugin list
        if not modname.startswith("_"):
            plugins.append(modname.replace("_", "-"))

    # we wrap the output depending on terminal size
    width = get_terminal_width()
    click.echo(f"Displaying plugins available for {base!r}")
    for line in format_output_in_columns(plugins, max_width=width):
        click.echo(line)
